@@ -5,7 +5,7 @@
 #
 # Copyright 2015-2016 VeriSign, Inc.
 #
-# Copyright 2016-2017 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -23,7 +23,6 @@
 
 from __future__ import unicode_literals
 
-import cgi
 import datetime
 
 # minimal support for python2.6
@@ -32,6 +31,12 @@ try:
 except ImportError:
     from ordereddict import OrderedDict
 
+# python3/python2 dual compatibility
+try:
+    from html import escape
+except ImportError:
+    from cgi import escape
+
 import dns.dnssec
 
 import dnsviz.format as fmt
@@ -58,6 +63,9 @@ class DomainNameAnalysisError(object):
             except KeyError:
                 raise TypeError('The "%s" keyword argument is required for instantiation.' % param)
 
+    def __hash__(self):
+        return id(self)
+
     def __str__(self):
         return self.code
 
@@ -83,16 +91,16 @@ class DomainNameAnalysisError(object):
 
     @property
     def html_description(self):
-        description_template_escaped = cgi.escape(self.description_template, True)
+        description_template_escaped = escape(self.description_template, True)
         template_kwargs_escaped = {}
         for n, v in self.template_kwargs.items():
             if isinstance(v, int):
                 template_kwargs_escaped[n] = v
             else:
                 if isinstance(v, str):
-                    template_kwargs_escaped[n] = cgi.escape(v)
+                    template_kwargs_escaped[n] = escape(v)
                 else:
-                    template_kwargs_escaped[n] = cgi.escape(str(v))
+                    template_kwargs_escaped[n] = escape(str(v))
         return description_template_escaped % template_kwargs_escaped
 
     def add_server_client(self, server, client, response):
@@ -301,7 +309,7 @@ class InceptionWithinClockSkew(RRSIGError):
     '''
     >>> e = InceptionWithinClockSkew(inception=datetime.datetime(2015,1,10,0,0,0), reference_time=datetime.datetime(2015,1,10,0,0,1))
     >>> e.description
-    'The value of the Signature Inception field of the RRSIG RR (2015-01-10 00:00:00) is within possible clock skew range of the current time (2015-01-10 00:00:01)'.
+    'The value of the Signature Inception field of the RRSIG RR (2015-01-10 00:00:00) is within possible clock skew range (1 second) of the current time (2015-01-10 00:00:01).'
     '''
 
     _abstract = False
@@ -319,7 +327,7 @@ class ExpirationWithinClockSkew(RRSIGError):
     '''
     >>> e = ExpirationWithinClockSkew(expiration=datetime.datetime(2015,1,10,0,0,1), reference_time=datetime.datetime(2015,1,10,0,0,0))
     >>> e.description
-    'The value of the Signature Expiration field of the RRSIG RR (2015-01-10 00:00:01) is within possible clock skew range of the current time (2015-01-10 00:00:00)'.
+    'The value of the Signature Expiration field of the RRSIG RR (2015-01-10 00:00:01) is within possible clock skew range (1 second) of the current time (2015-01-10 00:00:00).'
     '''
 
     _abstract = False
@@ -346,6 +354,97 @@ class SignatureInvalid(RRSIGError):
     references = ['RFC 4035, Sec. 5.3.3']
     required_params = []
 
+class RRSIGBadLength(RRSIGError):
+    pass
+
+class RRSIGBadLengthGOST(RRSIGBadLength):
+    '''
+    >>> e = RRSIGBadLengthGOST(length=500)
+    >>> e.description
+    'The length of the signature is 500 bits, but a GOST signature (DNSSEC algorithm 12) must be 512 bits long.'
+    '''
+    _abstract = False
+    description_template = 'The length of the signature is %(length)d bits, but a GOST signature (DNSSEC algorithm 12) must be 512 bits long.'
+    code = 'RRSIG_BAD_LENGTH_GOST'
+    references = ['RFC 5933, Sec. 5.2']
+    required_params = ['length']
+
+class RRSIGBadLengthECDSA(RRSIGBadLength):
+    curve = None
+    algorithm = None
+    correct_length = None
+    description_template = 'The length of the signature is %(length)d bits, but an ECDSA signature made with Curve %(curve)s (DNSSEC algorithm %(algorithm)d) must be %(correct_length)d bits long.'
+    references = ['RFC 6605, Sec. 4']
+    required_params = ['length']
+
+    def __init__(self, **kwargs):
+        super(RRSIGBadLengthECDSA, self).__init__(**kwargs)
+        self.template_kwargs['curve'] = self.curve
+        self.template_kwargs['algorithm'] = self.algorithm
+        self.template_kwargs['correct_length'] = self.correct_length
+
+class RRSIGBadLengthECDSA256(RRSIGBadLengthECDSA):
+    '''
+    >>> e = RRSIGBadLengthECDSA256(length=500)
+    >>> e.description
+    'The length of the signature is 500 bits, but an ECDSA signature made with Curve P-256 (DNSSEC algorithm 13) must be 512 bits long.'
+    '''
+    curve = 'P-256'
+    algorithm = 13
+    correct_length = 512
+    _abstract = False
+    code = 'RRSIG_BAD_LENGTH_ECDSA256'
+
+class RRSIGBadLengthECDSA384(RRSIGBadLengthECDSA):
+    '''
+    >>> e = RRSIGBadLengthECDSA384(length=500)
+    >>> e.description
+    'The length of the signature is 500 bits, but an ECDSA signature made with Curve P-384 (DNSSEC algorithm 14) must be 768 bits long.'
+    '''
+    curve = 'P-384'
+    algorithm = 14
+    correct_length = 768
+    _abstract = False
+    code = 'RRSIG_BAD_LENGTH_ECDSA384'
+
+class RRSIGBadLengthEdDSA(RRSIGBadLength):
+    curve = None
+    algorithm = None
+    correct_length = None
+    description_template = 'The length of the signature is %(length)d bits, but an %(curve)s signature (DNSSEC algorithm %(algorithm)d) must be %(correct_length)d bits long.'
+    references = ['RFC 8080, Sec. 4']
+    required_params = ['length']
+
+    def __init__(self, **kwargs):
+        super(RRSIGBadLengthEdDSA, self).__init__(**kwargs)
+        self.template_kwargs['curve'] = self.curve
+        self.template_kwargs['algorithm'] = self.algorithm
+        self.template_kwargs['correct_length'] = self.correct_length
+
+class RRSIGBadLengthEd25519(RRSIGBadLengthEdDSA):
+    '''
+    >>> e = RRSIGBadLengthEd25519(length=500)
+    >>> e.description
+    'The length of the signature is 500 bits, but an Ed25519 signature (DNSSEC algorithm 15) must be 512 bits long.'
+    '''
+    curve = 'Ed25519'
+    algorithm = 15
+    correct_length = 512
+    _abstract = False
+    code = 'RRSIG_BAD_LENGTH_ED25519'
+
+class RRSIGBadLengthEd448(RRSIGBadLengthEdDSA):
+    '''
+    >>> e = RRSIGBadLengthEd448(length=500)
+    >>> e.description
+    'The length of the signature is 500 bits, but an Ed448 signature (DNSSEC algorithm 16) must be 912 bits long.'
+    '''
+    curve = 'Ed448'
+    algorithm = 16
+    correct_length = 912
+    _abstract = False
+    code = 'RRSIG_BAD_LENGTH_ED448'
+
 class DSError(DomainNameAnalysisError):
     pass
 
@@ -1167,6 +1266,106 @@ class EDNSUndefinedFlagsSet(EDNSError):
     references = ['RFC 6891, Sec. 6.1.4']
     required_params = ['flags']
 
+class DNSCookieError(ResponseError):
+    pass
+
+class GratuitousCookie(DNSCookieError):
+    '''
+    >>> e = GratuitousCookie()
+    >>> e.description
+    'The server sent a COOKIE option when none was sent by the client.'
+    '''
+
+    _abstract = False
+    code = 'GRATUITOUS_COOKIE'
+    description_template = 'The server sent a COOKIE option when none was sent by the client.'
+    references = ['RFC 7873, Sec. 5.2.1']
+
+class MalformedCookieWithoutFORMERR(DNSCookieError):
+    '''
+    >>> e = MalformedCookieWithoutFORMERR()
+    >>> e.description
+    'The server appears to support DNS cookies but did not return a FORMERR status when issued a malformed COOKIE option.'
+    '''
+
+    _abstract = False
+    code = 'MALFORMED_COOKIE_WITHOUT_FORMERR'
+    description_template = 'The server appears to support DNS cookies but did not return a FORMERR status when issued a malformed COOKIE option.'
+    references = ['RFC 7873, Sec. 5.2.2']
+
+class NoCookieOption(DNSCookieError):
+    '''
+    >>> e = NoCookieOption()
+    >>> e.description
+    'The server appears to support DNS cookies but did not return a COOKIE option.'
+    '''
+
+    _abstract = False
+    code = 'NO_COOKIE_OPTION'
+    description_template = 'The server appears to support DNS cookies but did not return a COOKIE option.'
+    references = ['RFC 7873, Sec. 5.2.3']
+
+class NoServerCookieWithoutBADCOOKIE(DNSCookieError):
+    '''
+    >>> e = NoServerCookieWithoutBADCOOKIE()
+    >>> e.description
+    'The server appears to support DNS cookies but did not return a BADCOOKIE status when no server cookie was sent.'
+    '''
+
+    _abstract = False
+    code = 'NO_SERVER_COOKIE_WITHOUT_BADCOOKIE'
+    description_template = 'The server appears to support DNS cookies but did not return a BADCOOKIE status when no server cookie was sent.'
+    references = ['RFC 7873, Sec. 5.2.3']
+
+class InvalidServerCookieWithoutBADCOOKIE(DNSCookieError):
+    '''
+    >>> e = InvalidServerCookieWithoutBADCOOKIE()
+    >>> e.description
+    'The server appears to support DNS cookies but did not return a BADCOOKIE status when an invalid server cookie was sent.'
+    '''
+
+    _abstract = False
+    code = 'INVALID_SERVER_COOKIE_WITHOUT_BADCOOKIE'
+    description_template = 'The server appears to support DNS cookies but did not return a BADCOOKIE status when an invalid server cookie was sent.'
+    references = ['RFC 7873, Sec. 5.2.4']
+
+class NoServerCookie(DNSCookieError):
+    '''
+    >>> e = NoServerCookie()
+    >>> e.description
+    'The server appears to support DNS cookies but did not return a server cookie with its COOKIE option.'
+    '''
+
+    _abstract = False
+    code = 'NO_SERVER_COOKIE'
+    description_template = 'The server appears to support DNS cookies but did not return a server cookie with its COOKIE option.'
+    references = ['RFC 7873, Sec. 5.2.3']
+
+class ClientCookieMismatch(DNSCookieError):
+    '''
+    >>> e = ClientCookieMismatch()
+    >>> e.description
+    'The client cookie returned by the server did not match what was sent.'
+    '''
+
+    _abstract = False
+    code = 'CLIENT_COOKIE_MISMATCH'
+    description_template = 'The client cookie returned by the server did not match what was sent.'
+    references = ['RFC 7873, Sec. 5.3']
+
+class CookieInvalidLength(DNSCookieError):
+    '''
+    >>> e = CookieInvalidLength(length=61)
+    >>> e.description
+    'The cookie returned by the server had an invalid length of 61 bytes.'
+    '''
+
+    _abstract = False
+    code = 'COOKIE_INVALID_LENGTH'
+    description_template = 'The cookie returned by the server had an invalid length of %(length)d bytes.'
+    references = ['RFC 7873, Sec. 5.3']
+    required_params = ['length']
+
 class UnableToRetrieveDNSSECRecords(ResponseError):
     '''
     >>> e = UnableToRetrieveDNSSECRecords()
@@ -1404,6 +1603,59 @@ class PMTUExceeded(ResponseError):
         if self.template_kwargs['pmtu_lower_bound'] is not None and self.template_kwargs['pmtu_upper_bound'] is not None:
             self.template_kwargs['description'] += ' The PMTU was bounded between %(pmtu_lower_bound)d and %(pmtu_upper_bound)d bytes.' % self.template_kwargs
 
+class ForeignClassData(ResponseError):
+    section = None
+    description_template = 'Data of class %(cls)s was found in the %(section)s section of the response.'
+    references = ['RFC 1034', 'RFC 1035']
+    required_params = ['cls']
+
+    def __init__(self, **kwargs):
+        super(ForeignClassData, self).__init__(**kwargs)
+        self.template_kwargs['section'] = self.section
+
+class ForeignClassDataAnswer(ForeignClassData):
+    '''
+    >>> e = ForeignClassDataAnswer(cls='CH')
+    >>> e.description
+    'Data of class CH was found in the Answer section of the response.'
+    '''
+    section = 'Answer'
+    _abstract = False
+    code = 'FOREIGN_CLASS_DATA_ANSWER'
+
+class ForeignClassDataAuthority(ForeignClassData):
+    '''
+    >>> e = ForeignClassDataAuthority(cls='CH')
+    >>> e.description
+    'Data of class CH was found in the Authority section of the response.'
+    '''
+    section = 'Authority'
+    _abstract = False
+    code = 'FOREIGN_CLASS_DATA_AUTHORITY'
+
+class ForeignClassDataAdditional(ForeignClassData):
+    '''
+    >>> e = ForeignClassDataAdditional(cls='CH')
+    >>> e.description
+    'Data of class CH was found in the Additional section of the response.'
+    '''
+    section = 'Additional'
+    _abstract = False
+    code = 'FOREIGN_CLASS_DATA_ADDITIONAL'
+
+class CasePreservationError(ResponseError):
+    '''
+    >>> e = CasePreservationError(qname='ExAmPlE.CoM')
+    >>> e.description
+    'The case of the query name (ExAmPlE.CoM) was not preserved in the Question section of the response.'
+    '''
+
+    _abstract = False
+    code = 'CASE_NOT_PRESERVED'
+    description_template = '%(description)s'
+    description_template = 'The case of the query name (%(qname)s) was not preserved in the Question section of the response.'
+    required_params = ['qname']
+
 class DelegationError(DomainNameAnalysisError):
     pass
 
@@ -1756,6 +2008,97 @@ class DNSKEYNotAtZoneApex(DNSKEYError):
     code = 'DNSKEY_NOT_AT_ZONE_APEX'
     required_params = ['zone', 'name']
 
+class DNSKEYBadLength(DNSKEYError):
+    pass
+
+class DNSKEYBadLengthGOST(DNSKEYBadLength):
+    '''
+    >>> e = DNSKEYBadLengthGOST(length=500)
+    >>> e.description
+    'The length of the key is 500 bits, but a GOST public key (DNSSEC algorithm 12) must be 512 bits long.'
+    '''
+    _abstract = False
+    description_template = 'The length of the key is %(length)d bits, but a GOST public key (DNSSEC algorithm 12) must be 512 bits long.'
+    code = 'DNSKEY_BAD_LENGTH_GOST'
+    references = ['RFC 5933, Sec. 5.1']
+    required_params = ['length']
+
+class DNSKEYBadLengthECDSA(DNSKEYBadLength):
+    curve = None
+    algorithm = None
+    correct_length = None
+    description_template = 'The length of the key is %(length)d bits, but an ECDSA public key using Curve %(curve)s (DNSSEC algorithm %(algorithm)d) must be %(correct_length)d bits long.'
+    references = ['RFC 6605, Sec. 4']
+    required_params = ['length']
+
+    def __init__(self, **kwargs):
+        super(DNSKEYBadLengthECDSA, self).__init__(**kwargs)
+        self.template_kwargs['curve'] = self.curve
+        self.template_kwargs['algorithm'] = self.algorithm
+        self.template_kwargs['correct_length'] = self.correct_length
+
+class DNSKEYBadLengthECDSA256(DNSKEYBadLengthECDSA):
+    '''
+    >>> e = DNSKEYBadLengthECDSA256(length=500)
+    >>> e.description
+    'The length of the key is 500 bits, but an ECDSA public key using Curve P-256 (DNSSEC algorithm 13) must be 512 bits long.'
+    '''
+    curve = 'P-256'
+    algorithm = 13
+    correct_length = 512
+    _abstract = False
+    code = 'DNSKEY_BAD_LENGTH_ECDSA256'
+
+class DNSKEYBadLengthECDSA384(DNSKEYBadLengthECDSA):
+    '''
+    >>> e = DNSKEYBadLengthECDSA384(length=500)
+    >>> e.description
+    'The length of the key is 500 bits, but an ECDSA public key using Curve P-384 (DNSSEC algorithm 14) must be 768 bits long.'
+    '''
+    curve = 'P-384'
+    algorithm = 14
+    correct_length = 768
+    _abstract = False
+    code = 'DNSKEY_BAD_LENGTH_ECDSA384'
+
+class DNSKEYBadLengthEdDSA(DNSKEYBadLength):
+    curve = None
+    algorithm = None
+    correct_length = None
+    description_template = 'The length of the key is %(length)d bits, but an %(curve)s public key (DNSSEC algorithm %(algorithm)d) must be %(correct_length)d bits long.'
+    references = ['RFC 8080, Sec. 3']
+    required_params = ['length']
+
+    def __init__(self, **kwargs):
+        super(DNSKEYBadLengthEdDSA, self).__init__(**kwargs)
+        self.template_kwargs['curve'] = self.curve
+        self.template_kwargs['algorithm'] = self.algorithm
+        self.template_kwargs['correct_length'] = self.correct_length
+
+class DNSKEYBadLengthEd25519(DNSKEYBadLengthEdDSA):
+    '''
+    >>> e = DNSKEYBadLengthEd25519(length=500)
+    >>> e.description
+    'The length of the key is 500 bits, but an Ed25519 public key (DNSSEC algorithm 15) must be 256 bits long.'
+    '''
+    curve = 'Ed25519'
+    algorithm = 15
+    correct_length = 256
+    _abstract = False
+    code = 'DNSKEY_BAD_LENGTH_ED25519'
+
+class DNSKEYBadLengthEd448(DNSKEYBadLengthEdDSA):
+    '''
+    >>> e = DNSKEYBadLengthEd448(length=500)
+    >>> e.description
+    'The length of the key is 500 bits, but an Ed448 public key (DNSSEC algorithm 16) must be 456 bits long.'
+    '''
+    curve = 'Ed448'
+    algorithm = 16
+    correct_length = 456
+    _abstract = False
+    code = 'DNSKEY_BAD_LENGTH_ED448'
+
 class TrustAnchorError(DomainNameAnalysisError):
     pass
 
