@@ -1,8 +1,6 @@
 #
 # This file is a part of DNSViz, a tool suite for DNS/DNSSEC monitoring,
-# analysis, and visualization.  This file (or some portion thereof) is a
-# derivative work authored by VeriSign, Inc., and created in 2014, based on
-# code originally developed at Sandia National Laboratories.
+# analysis, and visualization.
 # Created by Casey Deccio (casey@deccio.net)
 #
 # Copyright 2012-2014 Sandia Corporation. Under the terms of Contract
@@ -11,6 +9,8 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
+# Copyright 2016-2019 Casey Deccio
+#
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
@@ -28,7 +28,6 @@
 from __future__ import unicode_literals
 
 import base64
-import cgi
 import datetime
 import logging
 
@@ -38,6 +37,12 @@ try:
 except ImportError:
     from ordereddict import OrderedDict
 
+# python3/python2 dual compatibility
+try:
+    from html import escape
+except ImportError:
+    from cgi import escape
+
 import dns.name, dns.rdatatype
 
 from dnsviz import base32
@@ -153,6 +158,15 @@ dname_status_mapping = {
         DNAME_STATUS_INVALID: 'INVALID',
 }
 
+RRSIG_SIG_LENGTHS_BY_ALGORITHM = {
+        12: 512, 13: 512, 14: 768, 15: 512, 16: 912,
+}
+RRSIG_SIG_LENGTH_ERRORS = {
+        12: Errors.RRSIGBadLengthGOST, 13: Errors.RRSIGBadLengthECDSA256,
+        14: Errors.RRSIGBadLengthECDSA384, 15: Errors.RRSIGBadLengthEd25519,
+        16: Errors.RRSIGBadLengthEd448,
+}
+
 class RRSIGStatus(object):
     def __init__(self, rrset, rrsig, dnskey, zone_name, reference_ts, supported_algs):
         self.rrset = rrset
@@ -206,6 +220,11 @@ class RRSIGStatus(object):
                 if self.validation_status == RRSIG_STATUS_VALID:
                     self.validation_status = RRSIG_STATUS_INVALID
 
+        sig_len = len(self.rrsig.signature) << 3
+        if self.rrsig.algorithm in RRSIG_SIG_LENGTHS_BY_ALGORITHM and \
+                sig_len != RRSIG_SIG_LENGTHS_BY_ALGORITHM[self.rrsig.algorithm]:
+            self.errors.append(RRSIG_SIG_LENGTH_ERRORS[self.rrsig.algorithm](length=sig_len))
+
         if self.reference_ts < self.rrsig.inception:
             if self.validation_status == RRSIG_STATUS_VALID:
                 self.validation_status = RRSIG_STATUS_PREMATURE
@@ -243,7 +262,7 @@ class RRSIGStatus(object):
                 erroneous_status
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -382,7 +401,7 @@ class DSStatus(object):
                 erroneous_status
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -535,7 +554,7 @@ class NSECStatusNXDOMAIN(NSECStatus):
                 (erroneous_status or nsec_list)
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -786,7 +805,7 @@ class NSECStatusNODATA(NSECStatus):
                 (erroneous_status or nsec_list)
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -999,7 +1018,7 @@ class NSEC3StatusNXDOMAIN(NSEC3Status):
                 (erroneous_status or nsec3_list)
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -1350,7 +1369,7 @@ class NSEC3StatusNODATA(NSEC3Status):
                 (erroneous_status or nsec3_list)
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -1484,7 +1503,7 @@ class CNAMEFromDNAMEStatus(object):
                 (erroneous_status or dname_serialized)
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
