@@ -6,7 +6,7 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
-# Copyright 2016-2017 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -33,6 +33,12 @@ import os
 import re
 import sys
 
+# minimal support for python2.6
+try:
+    from collections import OrderedDict
+except ImportError:
+    from ordereddict import OrderedDict
+
 import dns.exception, dns.name
 
 from dnsviz.analysis import OfflineDomainNameAnalysis, DNS_RAW_VERSION
@@ -55,29 +61,37 @@ except ImportError:
 LOCAL_MEDIA_URL = 'file://' + DNSVIZ_SHARE_PATH
 DNSSEC_TEMPLATE_FILE = os.path.join(DNSVIZ_SHARE_PATH, 'html', 'dnssec-template.html')
 
-logger = logging.getLogger('dnsviz.analysis.offline')
+logging.basicConfig(level=logging.WARNING, format='%(message)s')
+logger = logging.getLogger()
 
 def usage(err=None):
     if err is not None:
         err += '\n\n'
     else:
         err = ''
-    sys.stderr.write('''%sUsage: dnsviz graph [options] [domain name...]
-Options:
-    -f <filename>  - read names from a file
-    -r <filename>  - read diagnostic queries from a file
-    -t <filename>  - specify file containing trusted keys
-    -R <type>[,<type>...]
-                   - Process queries of only the specified type(s)
-    -O             - derive the filename(s) from the format and domain name(s)
-    -o <filename>  - save the output to the specified file
-    -T <format>    - specify the format of the output
-    -h             - display the usage and exit
-''' % (err))
+    sys.stderr.write('''%sUsage: %s %s [options] [domain_name...]
 
-def finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, filename):
+Graph the assessment of diagnostic DNS queries.
+
+Options:
+    -f <filename>  - Read names from a file.
+    -r <filename>  - Read diagnostic queries from a file.
+    -t <filename>  - Use trusted keys from the designated file.
+    -C             - Enforce DNS cookies strictly.
+    -R <type>[,<type>...]
+                   - Process queries of only the specified type(s).
+    -e             - Do not remove redundant RRSIG edges from the graph.
+    -O             - Derive the filename(s) from the format and domain name(s).
+    -o <filename>  - Save the output to the specified file.
+    -T <format>    - Use the specified output format.
+    -h             - Display the usage and exit.
+''' % (err, sys.argv[0], __name__.split('.')[-1]))
+
+def finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, filename, remove_edges):
     G.add_trust(trusted_keys)
-    G.remove_extra_edges()
+
+    if remove_edges:
+        G.remove_extra_edges()
 
     if fmt == 'html':
         try:
@@ -87,7 +101,8 @@ def finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, filename):
             sys.exit(3)
 
         try:
-            template_str = io.open(DNSSEC_TEMPLATE_FILE, 'r', encoding='utf-8').read()
+            with io.open(DNSSEC_TEMPLATE_FILE, 'r', encoding='utf-8') as fh:
+                template_str = fh.read()
         except IOError as e:
             logger.error('Error reading template file "%s": %s' % (DNSSEC_TEMPLATE_FILE, e.strerror))
             sys.exit(3)
@@ -119,12 +134,6 @@ def finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, filename):
                     logger.error(str(e))
                 sys.exit(3)
 
-def test_m2crypto():
-    try:
-        import M2Crypto
-    except ImportError:
-        sys.stderr.write('''Warning: M2Crypto is not installed; cryptographic validation of signatures and digests will not be available.\n''')
-
 def test_pygraphviz():
     try:
         from pygraphviz import release
@@ -133,24 +142,23 @@ def test_pygraphviz():
             major = int(major)
             minor = int(re.sub(r'(\d+)[^\d].*', r'\1', minor))
             if (major, minor) < (1,1):
-                sys.stderr.write('''pygraphviz version >= 1.1 is required, but version %s is installed.\n''' % release.version)
+                logger.error('''pygraphviz version >= 1.1 is required, but version %s is installed.''' % release.version)
                 sys.exit(2)
         except ValueError:
-            sys.stderr.write('''pygraphviz version >= 1.1 is required, but version %s is installed.\n''' % release.version)
+            logger.error('''pygraphviz version >= 1.1 is required, but version %s is installed.''' % release.version)
             sys.exit(2)
     except ImportError:
-        sys.stderr.write('''pygraphviz is required, but not installed.\n''')
+        logger.error('''pygraphviz is required, but not installed.''')
         sys.exit(2)
 
 def main(argv):
     try:
-        test_m2crypto()
         test_pygraphviz()
 
         try:
-            opts, args = getopt.getopt(argv[1:], 'f:r:R:t:Oo:T:h')
+            opts, args = getopt.getopt(argv[1:], 'f:r:R:et:COo:T:h')
         except getopt.GetoptError as e:
-            usage(str(e))
+            sys.stderr.write('%s\n' % str(e))
             sys.exit(1)
 
         # collect trusted keys
@@ -158,14 +166,15 @@ def main(argv):
         for opt, arg in opts:
             if opt == '-t':
                 try:
-                    tk_str = io.open(arg, 'r', encoding='utf-8').read()
+                    with io.open(arg, 'r', encoding='utf-8') as fh:
+                        tk_str = fh.read()
                 except IOError as e:
-                    sys.stderr.write('%s: "%s"\n' % (e.strerror, arg))
+                    logger.error('%s: "%s"' % (e.strerror, arg))
                     sys.exit(3)
                 try:
                     trusted_keys.extend(get_trusted_keys(tk_str))
                 except dns.exception.DNSException:
-                    sys.stderr.write('There was an error parsing the trusted keys file: "%s"\n' % arg)
+                    logger.error('There was an error parsing the trusted keys file: "%s"' % arg)
                     sys.exit(3)
 
         opts = dict(opts)
@@ -174,23 +183,27 @@ def main(argv):
             sys.exit(0)
 
         if '-f' in opts and args:
-            usage('If -f is used, then domain names may not supplied as command line arguments.')
+            sys.stderr.write('If -f is used, then domain names may not supplied as command line arguments.\n')
             sys.exit(1)
 
         if '-R' in opts:
             try:
                 rdtypes = opts['-R'].split(',')
             except ValueError:
-                usage('The list of types was invalid: "%s"' % opts['-R'])
+                sys.stderr.write('The list of types was invalid: "%s"\n' % opts['-R'])
                 sys.exit(1)
             try:
                 rdtypes = [dns.rdatatype.from_text(x) for x in rdtypes]
             except dns.rdatatype.UnknownRdatatype:
-                usage('The list of types was invalid: "%s"' % opts['-R'])
+                sys.stderr.write('The list of types was invalid: "%s"\n' % opts['-R'])
                 sys.exit(1)
         else:
             rdtypes = None
 
+        strict_cookies = '-C' in opts
+
+        remove_edges = '-e' not in opts
+
         if '-T' in opts:
             fmt = opts['-T']
         elif '-o' in opts:
@@ -198,31 +211,31 @@ def main(argv):
         else:
             fmt = 'dot'
         if fmt not in ('dot','png','jpg','svg','html'):
-            usage('Image format unrecognized: "%s"' % fmt)
+            sys.stderr.write('Image format unrecognized: "%s"\n' % fmt)
             sys.exit(1)
 
         if '-o' in opts and '-O' in opts:
-            usage('The -o and -O options may not be used together.')
+            sys.stderr.write('The -o and -O options may not be used together.\n')
             sys.exit(1)
 
-        handler = logging.StreamHandler()
-        handler.setLevel(logging.WARNING)
-        logger.addHandler(handler)
-        logger.setLevel(logging.WARNING)
-
         if '-r' not in opts or opts['-r'] == '-':
             opt_r = sys.stdin.fileno()
         else:
             opt_r = opts['-r']
         try:
-            analysis_str = io.open(opt_r, 'r', encoding='utf-8').read()
+            with io.open(opt_r, 'r', encoding='utf-8') as fh:
+                analysis_str = fh.read()
         except IOError as e:
             logger.error('%s: "%s"' % (e.strerror, opts.get('-r', '-')))
             sys.exit(3)
+        if not analysis_str:
+            if opt_r != sys.stdin.fileno():
+                logger.error('No input.')
+            sys.exit(3)
         try:
             analysis_structured = json.loads(analysis_str)
         except ValueError:
-            logger.error('There was an error parsing the json input: "%s"' % opts.get('-r', '-'))
+            logger.error('There was an error parsing the JSON input: "%s"' % opts.get('-r', '-'))
             sys.exit(3)
 
         # check version
@@ -241,7 +254,7 @@ def main(argv):
             logger.error('Version %d.%d of JSON input is incompatible with this software.' % (major_vers, minor_vers))
             sys.exit(3)
 
-        names = []
+        names = OrderedDict()
         if '-f' in opts:
             if opts['-f'] == '-':
                 opts['-f'] = sys.stdin.fileno()
@@ -259,7 +272,8 @@ def main(argv):
                 except dns.exception.DNSException:
                     logger.error('The domain name was invalid: "%s"' % name)
                 else:
-                    names.append(name)
+                    if name not in names:
+                        names[name] = None
             f.close()
         else:
             if args:
@@ -270,7 +284,7 @@ def main(argv):
                 try:
                     args = analysis_structured['_meta._dnsviz.']['names']
                 except KeyError:
-                    logger.error('No names found in json input!')
+                    logger.error('No names found in JSON input!')
                     sys.exit(3)
             for name in args:
                 try:
@@ -280,7 +294,8 @@ def main(argv):
                 except dns.exception.DNSException:
                     logger.error('The domain name was invalid: "%s"' % name)
                 else:
-                    names.append(name)
+                    if name not in names:
+                        names[name] = None
 
         latest_analysis_date = None
         name_objs = []
@@ -290,7 +305,7 @@ def main(argv):
             if name_str not in analysis_structured or analysis_structured[name_str].get('stub', True):
                 logger.error('The analysis of "%s" was not found in the input.' % lb2s(name.to_text()))
                 continue
-            name_obj = OfflineDomainNameAnalysis.deserialize(name, analysis_structured, cache)
+            name_obj = OfflineDomainNameAnalysis.deserialize(name, analysis_structured, cache, strict_cookies=strict_cookies)
             name_objs.append(name_obj)
 
             if latest_analysis_date is None or latest_analysis_date > name_obj.analysis_end:
@@ -328,14 +343,14 @@ def main(argv):
                     name = 'root'
                 else:
                     name = lb2s(name_obj.name.canonicalize().to_text()).rstrip('.')
-                finish_graph(G, [name_obj], rdtypes, trusted_keys, fmt, '%s.%s' % (name, fmt))
+                finish_graph(G, [name_obj], rdtypes, trusted_keys, fmt, '%s.%s' % (name, fmt), remove_edges)
                 G = DNSAuthGraph()
 
         if '-O' not in opts:
             if '-o' not in opts or opts['-o'] == '-':
-                finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, None)
+                finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, None, remove_edges)
             else:
-                finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, opts['-o'])
+                finish_graph(G, name_objs, rdtypes, trusted_keys, fmt, opts['-o'], remove_edges)
 
     except KeyboardInterrupt:
         logger.error('Interrupted.')
