@@ -6,7 +6,7 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
-# Copyright 2016-2017 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -57,7 +57,8 @@ except ImportError:
     else:
         raise
 
-logger = logging.getLogger('dnsviz.analysis.offline')
+logging.basicConfig(level=logging.WARNING, format='%(message)s')
+logger = logging.getLogger()
 
 TERM_COLOR_MAP = {
     'BOLD': '\033[1m',
@@ -98,16 +99,20 @@ def usage(err=None):
         err += '\n\n'
     else:
         err = ''
-    sys.stderr.write('''%sUsage: dnsviz grok [options] [domain name...]
+    sys.stderr.write('''%sUsage: %s %s [options] [domain_name...]
+
+Assess diagnostic DNS queries.
+
 Options:
-    -f <filename>  - read names from a file
-    -r <filename>  - read diagnostic queries from a file
-    -t <filename>  - specify file containing trusted keys
-    -o <filename>  - save the output to the specified file
-    -c             - make json output minimal instead of pretty
-    -l <loglevel>  - set log level to one of: error, warning, info, debug
-    -h             - display the usage and exit
-''' % (err))
+    -f <filename>  - Read names from a file.
+    -r <filename>  - Read diagnostic queries from a file.
+    -t <filename>  - Use trusted keys from the designated file.
+    -C             - Enforce DNS cookies strictly.
+    -o <filename>  - Save the output to the specified file.
+    -c             - Format JSON output minimally, instead of "pretty".
+    -l <loglevel>  - Log at the specified level: error, warning, info, debug.
+    -h             - Display the usage and exit.
+''' % (err, sys.argv[0], __name__.split('.')[-1]))
 
 def color_json(s):
     error = None
@@ -145,12 +150,6 @@ def color_json(s):
 
     return s1.rstrip()
 
-def test_m2crypto():
-    try:
-        import M2Crypto
-    except ImportError:
-        sys.stderr.write('''Warning: M2Crypto is not installed; cryptographic validation of signatures and digests will not be available.\n''')
-
 def test_pygraphviz():
     try:
         from pygraphviz import release
@@ -159,24 +158,21 @@ def test_pygraphviz():
             major = int(major)
             minor = int(re.sub(r'(\d+)[^\d].*', r'\1', minor))
             if (major, minor) < (1,1):
-                sys.stderr.write('''pygraphviz version >= 1.1 is required, but version %s is installed.\n''' % release.version)
+                logger.error('''pygraphviz version >= 1.1 is required, but version %s is installed.''' % release.version)
                 sys.exit(2)
         except ValueError:
-            sys.stderr.write('''pygraphviz version >= 1.1 is required, but version %s is installed.\n''' % release.version)
+            logger.error('''pygraphviz version >= 1.1 is required, but version %s is installed.''' % release.version)
             sys.exit(2)
     except ImportError:
-        sys.stderr.write('''pygraphviz is required, but not installed.\n''')
+        logger.error('''pygraphviz is required, but not installed.''')
         sys.exit(2)
 
 def main(argv):
     try:
-        test_m2crypto()
-
-        #TODO remove -p option (it is now the default, and -c is used to change it)
         try:
-            opts, args = getopt.getopt(argv[1:], 'f:r:t:o:cpl:h')
+            opts, args = getopt.getopt(argv[1:], 'f:r:t:Co:cl:h')
         except getopt.GetoptError as e:
-            usage(str(e))
+            sys.stderr.write('%s\n' % str(e))
             sys.exit(1)
 
         # collect trusted keys
@@ -184,14 +180,15 @@ def main(argv):
         for opt, arg in opts:
             if opt == '-t':
                 try:
-                    tk_str = io.open(arg, 'r', encoding='utf-8').read()
+                    with io.open(arg, 'r', encoding='utf-8') as fh:
+                        tk_str = fh.read()
                 except IOError as e:
-                    sys.stderr.write('%s: "%s"\n' % (e.strerror, arg))
+                    logger.error('%s: "%s"' % (e.strerror, arg))
                     sys.exit(3)
                 try:
                     trusted_keys.extend(get_trusted_keys(tk_str))
                 except dns.exception.DNSException:
-                    sys.stderr.write('There was an error parsing the trusted keys file: "%s"\n' % arg)
+                    logger.error('There was an error parsing the trusted keys file: "%s"' % arg)
                     sys.exit(3)
 
         opts = dict(opts)
@@ -200,7 +197,7 @@ def main(argv):
             sys.exit(0)
 
         if '-f' in opts and args:
-            usage('If -f is used, then domain names may not supplied as command line arguments.')
+            sys.stderr.write('If -f is used, then domain names may not supplied as command line arguments.\n')
             sys.exit(1)
 
         if '-l' in opts:
@@ -213,28 +210,31 @@ def main(argv):
             elif opts['-l'] == 'debug':
                 loglevel = logging.DEBUG
             else:
-                usage('Invalid log level: "%s"' % opts['-l'])
+                sys.stderr.write('Invalid log level: "%s"\n' % opts['-l'])
                 sys.exit(1)
         else:
             loglevel = logging.DEBUG
-        handler = logging.StreamHandler()
-        handler.setLevel(logging.WARNING)
-        logger.addHandler(handler)
-        logger.setLevel(logging.WARNING)
+
+        strict_cookies = '-C' in opts
 
         if '-r' not in opts or opts['-r'] == '-':
             opt_r = sys.stdin.fileno()
         else:
             opt_r = opts['-r']
         try:
-            analysis_str = io.open(opt_r, 'r', encoding='utf-8').read()
+            with io.open(opt_r, 'r', encoding='utf-8') as fh:
+                analysis_str = fh.read()
         except IOError as e:
             logger.error('%s: "%s"' % (e.strerror, opts.get('-r', '-')))
             sys.exit(3)
+        if not analysis_str:
+            if opt_r != sys.stdin.fileno():
+                logger.error('No input.')
+            sys.exit(3)
         try:
             analysis_structured = json.loads(analysis_str)
         except ValueError:
-            logger.error('There was an error parsing the json input: "%s"' % opts.get('-r', '-'))
+            logger.error('There was an error parsing the JSON input: "%s"' % opts.get('-r', '-'))
             sys.exit(3)
 
         # check version
@@ -253,7 +253,7 @@ def main(argv):
             logger.error('Version %d.%d of JSON input is incompatible with this software.' % (major_vers, minor_vers))
             sys.exit(3)
 
-        names = []
+        names = OrderedDict()
         if '-f' in opts:
             if opts['-f'] == '-':
                 opts['-f'] = sys.stdin.fileno()
@@ -271,7 +271,8 @@ def main(argv):
                 except dns.exception.DNSException:
                     logger.error('The domain name was invalid: "%s"' % name)
                 else:
-                    names.append(name)
+                    if name not in names:
+                        names[name] = None
             f.close()
         else:
             if args:
@@ -282,7 +283,7 @@ def main(argv):
                 try:
                     args = analysis_structured['_meta._dnsviz.']['names']
                 except KeyError:
-                    logger.error('No names found in json input!')
+                    logger.error('No names found in JSON input!')
                     sys.exit(3)
             for name in args:
                 try:
@@ -292,7 +293,8 @@ def main(argv):
                 except dns.exception.DNSException:
                     logger.error('The domain name was invalid: "%s"' % name)
                 else:
-                    names.append(name)
+                    if name not in names:
+                        names[name] = None
 
         if '-o' not in opts or opts['-o'] == '-':
             opts['-o'] = sys.stdout.fileno()
@@ -318,7 +320,7 @@ def main(argv):
             if name_str not in analysis_structured or analysis_structured[name_str].get('stub', True):
                 logger.error('The analysis of "%s" was not found in the input.' % lb2s(name.to_text()))
                 continue
-            name_obj = OfflineDomainNameAnalysis.deserialize(name, analysis_structured, cache)
+            name_obj = OfflineDomainNameAnalysis.deserialize(name, analysis_structured, cache, strict_cookies=strict_cookies)
             name_objs.append(name_obj)
 
         if not name_objs:
