@@ -6,7 +6,7 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
-# Copyright 2016-2017 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,23 +25,24 @@
 from __future__ import unicode_literals
 
 import atexit
+import binascii
 import codecs
 import errno
 import getopt
 import io
 import json
 import logging
-import os
-import re
-import signal
-import socket
-import sys
 import multiprocessing
 import multiprocessing.managers
-import signal
+import os
+import random
+import re
 import shutil
+import signal
+import socket
 import struct
 import subprocess
+import sys
 import tempfile
 import threading
 import time
@@ -62,19 +63,21 @@ else:
 
 import dns.edns, dns.exception, dns.message, dns.name, dns.rdata, dns.rdataclass, dns.rdatatype, dns.rdtypes.ANY.NS, dns.rdtypes.IN.A, dns.rdtypes.IN.AAAA, dns.resolver, dns.rrset
 
-from dnsviz.analysis import WILDCARD_EXPLICIT_DELEGATION, PrivateAnalyst, PrivateRecursiveAnalyst, OnlineDomainNameAnalysis, NetworkConnectivityException, DNS_RAW_VERSION
+from dnsviz.analysis import COOKIE_STANDIN, WILDCARD_EXPLICIT_DELEGATION, PrivateAnalyst, PrivateRecursiveAnalyst, OnlineDomainNameAnalysis, NetworkConnectivityException, DNS_RAW_VERSION
 import dnsviz.format as fmt
 from dnsviz.ipaddr import IPAddr
-from dnsviz.query import StandardRecursiveQueryCD
+from dnsviz.query import DiagnosticQuery, QuickDNSSECQuery, StandardRecursiveQueryCD
 from dnsviz.resolver import DNSAnswer, Resolver, PrivateFullResolver
 from dnsviz import transport
 from dnsviz.util import get_client_address, get_root_hints
 lb2s = fmt.latin1_binary_to_string
 
-logger = logging.getLogger('dnsviz.analysis.online')
+logging.basicConfig(level=logging.WARNING, format='%(message)s')
+logger = logging.getLogger()
 
 # this needs to be global because of multiprocessing
 tm = None
+th_factories = None
 resolver = None
 bootstrap_resolver = None
 explicit_delegations = None
@@ -118,11 +121,14 @@ def _init_stub_resolver():
 def _init_full_resolver():
     global resolver
 
+    quick_query = QuickDNSSECQuery.add_mixin(CustomQueryMixin).add_server_cookie(COOKIE_STANDIN)
+    diagnostic_query = DiagnosticQuery.add_mixin(CustomQueryMixin).add_server_cookie(COOKIE_STANDIN)
+
     # now that we have the hints, make resolver a full resolver instead of a stub
     hints = get_root_hints()
     for key in explicit_delegations:
         hints[key] = explicit_delegations[key]
-    resolver = PrivateFullResolver(hints, odd_ports=odd_ports, transport_manager=tm)
+    resolver = PrivateFullResolver(hints, query_cls=(quick_query, diagnostic_query), odd_ports=odd_ports, cookie_standin=COOKIE_STANDIN, transport_manager=tm)
 
 def _init_interrupt_handler():
     signal.signal(signal.SIGINT, _raise_eof)
@@ -136,14 +142,14 @@ def _init_subprocess(use_full):
     _init_interrupt_handler()
 
 def _analyze(args):
-    (cls, name, dlv_domain, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, \
-            stop_at_explicit, extra_rdtypes, explicit_only, cache, cache_level, cache_lock, th_factories) = args
+    (cls, name, rdclass, dlv_domain, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, \
+            stop_at_explicit, extra_rdtypes, explicit_only, cache, cache_level, cache_lock) = args
     if ceiling is not None and name.is_subdomain(ceiling):
         c = ceiling
     else:
         c = name
     try:
-        a = cls(name, dlv_domain=dlv_domain, try_ipv4=try_ipv4, try_ipv6=try_ipv6, client_ipv4=client_ipv4, client_ipv6=client_ipv6, query_class_mixin=query_class_mixin, ceiling=c, edns_diagnostics=edns_diagnostics, explicit_delegations=explicit_delegations, stop_at_explicit=stop_at_explicit, odd_ports=odd_ports, extra_rdtypes=extra_rdtypes, explicit_only=explicit_only, analysis_cache=cache, cache_level=cache_level, analysis_cache_lock=cache_lock, transport_manager=tm, th_factories=th_factories, resolver=resolver)
+        a = cls(name, rdclass=rdclass, dlv_domain=dlv_domain, try_ipv4=try_ipv4, try_ipv6=try_ipv6, client_ipv4=client_ipv4, client_ipv6=client_ipv6, query_class_mixin=query_class_mixin, ceiling=c, edns_diagnostics=edns_diagnostics, explicit_delegations=explicit_delegations, stop_at_explicit=stop_at_explicit, odd_ports=odd_ports, extra_rdtypes=extra_rdtypes, explicit_only=explicit_only, analysis_cache=cache, cache_level=cache_level, analysis_cache_lock=cache_lock, transport_manager=tm, th_factories=th_factories, resolver=resolver)
         return a.analyze()
     # re-raise a KeyboardInterrupt, as this means we've been interrupted
     except KeyboardInterrupt:
@@ -159,14 +165,12 @@ def _analyze(args):
         logger.exception('Error analyzing %s' % fmt.humanize_name(name))
         return None
 
-class CustomQueryMixin(object):
-    pass
-
 class BulkAnalyst(object):
     analyst_cls = PrivateAnalyst
     use_full_resolver = True
 
-    def __init__(self, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, extra_rdtypes, explicit_only, dlv_domain, th_factories):
+    def __init__(self, rdclass, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, extra_rdtypes, explicit_only, dlv_domain):
+        self.rdclass = rdclass
         self.try_ipv4 = try_ipv4
         self.try_ipv6 = try_ipv6
         self.client_ipv4 = client_ipv4
@@ -179,14 +183,13 @@ class BulkAnalyst(object):
         self.extra_rdtypes = extra_rdtypes
         self.explicit_only = explicit_only
         self.dlv_domain = dlv_domain
-        self.th_factories = th_factories
 
         self.cache = {}
         self.cache_lock = threading.Lock()
 
     def _name_to_args_iter(self, names):
         for name in names:
-            yield (self.analyst_cls, name, self.dlv_domain, self.try_ipv4, self.try_ipv6, self.client_ipv4, self.client_ipv6, self.query_class_mixin, self.ceiling, self.edns_diagnostics, self.stop_at_explicit, self.extra_rdtypes, self.explicit_only, self.cache, self.cache_level, self.cache_lock, self.th_factories)
+            yield (self.analyst_cls, name, self.rdclass, self.dlv_domain, self.try_ipv4, self.try_ipv6, self.client_ipv4, self.client_ipv6, self.query_class_mixin, self.ceiling, self.edns_diagnostics, self.stop_at_explicit, self.extra_rdtypes, self.explicit_only, self.cache, self.cache_level, self.cache_lock)
 
     def analyze(self, names, flush_func=None):
         name_objs = []
@@ -276,8 +279,8 @@ class ParallelAnalystMixin(object):
     analyst_cls = MultiProcessAnalyst
     use_full_resolver = None
 
-    def __init__(self, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, extra_rdtypes, explicit_only, dlv_domain, th_factories, processes):
-        super(ParallelAnalystMixin, self).__init__(try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, extra_rdtypes, explicit_only, dlv_domain, th_factories)
+    def __init__(self, rdclass, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, extra_rdtypes, explicit_only, dlv_domain, processes):
+        super(ParallelAnalystMixin, self).__init__(rdclass, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, extra_rdtypes, explicit_only, dlv_domain)
         self.manager = multiprocessing.managers.SyncManager()
         self.manager.start()
 
@@ -342,7 +345,8 @@ def name_addr_mappings_from_string(domain, addr_mappings, delegation_mapping, re
             if require_name:
                 mappings_from_file = []
                 try:
-                    s = io.open(mapping, 'r', encoding='utf-8').read()
+                    with io.open(mapping, 'r', encoding='utf-8') as fh:
+                        s = fh.read()
                 except IOError as e:
                     usage('%s: "%s"' % (e.strerror, mapping))
                     sys.exit(3)
@@ -507,7 +511,8 @@ def ds_from_string(domain, dss, delegation_mapping):
         # if the value is actually a path, then check it as a zone file
         if os.path.isfile(ds):
             try:
-                s = io.open(ds, 'r', encoding='utf-8').read()
+                with io.open(ds, 'r', encoding='utf-8') as fh:
+                    s = fh.read()
             except IOError as e:
                 usage('%s: "%s"' % (e.strerror, ds))
                 sys.exit(3)
@@ -561,11 +566,12 @@ def _cleanup_process(working_dir, pid):
 
 def _serve_zone(zone, zone_file, port):
     tmpdir = tempfile.mkdtemp(prefix='dnsviz')
+    env = { 'PATH': '%s:/sbin:/usr/sbin:/usr/local/sbin' % (os.environ.get('PATH', '')) }
     pid = None
 
     io.open('%s/named.conf' % tmpdir, 'w', encoding='utf-8').write('''
 options {
-    directory "%s";
+	directory "%s";
 	pid-file "named.pid";
 	listen-on port %s { localhost; };
 	listen-on-v6 port %s { localhost; };
@@ -585,7 +591,7 @@ logging {
 ''' % (tmpdir, port, port, lb2s(zone.to_text()), os.path.abspath(zone_file), tmpdir))
 
     try:
-        p = subprocess.Popen(['named-checkconf', '-z', '%s/named.conf' % tmpdir], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+        p = subprocess.Popen(['named-checkconf', '-z', '%s/named.conf' % tmpdir], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
     except OSError as e:
         usage('This option requires executing named-checkconf.  Please ensure that it is installed and in PATH (%s).' % e)
         _cleanup_process(tmpdir, pid)
@@ -598,7 +604,7 @@ logging {
         sys.exit(1)
 
     try:
-        p = subprocess.Popen(['named', '-L', '%s/named.log' % tmpdir, '-c', '%s/named.conf' % tmpdir], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+        p = subprocess.Popen(['named', '-c', '%s/named.conf' % tmpdir], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
     except OSError as e:
         usage('This option requires executing named.  Please ensure that it is installed and in PATH (%s).' % e)
         _cleanup_process(tmpdir, pid)
@@ -607,7 +613,8 @@ logging {
     (stdout, stderr) = p.communicate()
     if p.returncode != 0:
         try:
-            log = io.open('%s/named.log' % tmpdir, 'r', encoding='utf-8').read()
+            with io.open('%s/named.log' % tmpdir, 'r', encoding='utf-8') as fh:
+                log = fh.read()
         except IOError as e:
             log = ''
         if not log:
@@ -617,7 +624,8 @@ logging {
         sys.exit(1)
 
     try:
-        pid = int(io.open('%s/named.pid' % tmpdir, 'r', encoding='utf-8').read())
+        with io.open('%s/named.pid' % tmpdir, 'r', encoding='utf-8') as fh:
+            pid = int(fh.read())
     except (IOError, ValueError) as e:
         usage('There was an error detecting the process ID for named: %s' % e)
         _cleanup_process(tmpdir, pid)
@@ -627,10 +635,10 @@ logging {
 
 def _get_ecs_option(s):
     try:
-        addr, prefix = s.split('/', 1)
+        addr, prefix_len = s.split('/', 1)
     except ValueError:
         addr = s
-        prefix = None
+        prefix_len = None
 
     try:
         addr = IPAddr(addr)
@@ -645,26 +653,26 @@ def _get_ecs_option(s):
         addrlen = 16
         family = 2
 
-    if prefix is None:
-        prefix = addrlen << 3
+    if prefix_len is None:
+        prefix_len = addrlen << 3
     else:
         try:
-            prefix = int(prefix)
+            prefix_len = int(prefix_len)
         except ValueError:
-            usage('The mask length was invalid: "%s"' % prefix)
+            usage('The mask length was invalid: "%s"' % prefix_len)
             sys.exit(1)
 
-        if prefix < 0 or prefix > (addrlen << 3):
-            usage('The mask length was invalid: "%d"' % prefix)
+        if prefix_len < 0 or prefix_len > (addrlen << 3):
+            usage('The mask length was invalid: "%d"' % prefix_len)
             sys.exit(1)
 
-    bytes_masked, remainder = divmod(prefix, 8)
+    bytes_masked, remainder = divmod(prefix_len, 8)
     if remainder:
         bytes_masked += 1
 
-    wire = struct.pack('!H', family)
-    wire += struct.pack('!B', prefix)
-    wire += struct.pack('!B', 0)
+    wire = struct.pack(b'!H', family)
+    wire += struct.pack(b'!B', prefix_len)
+    wire += struct.pack(b'!B', 0)
     wire += addr._ipaddr_bytes[:bytes_masked]
 
     return dns.edns.GenericOption(8, wire)
@@ -673,45 +681,74 @@ def _get_nsid_option():
 
     return dns.edns.GenericOption(dns.edns.NSID, b'')
 
+def _get_dns_cookie_option(cookie=None):
+    if cookie is None:
+        r = random.getrandbits(64)
+        cookie = struct.pack(b'Q', r)
+    else:
+        try:
+            cookie = binascii.unhexlify(cookie)
+        except TypeError:
+            usage('The DNS cookie provided was not valid hexadecimal: "%s"' % cookie)
+            sys.exit(1)
+
+        if len(cookie) != 8:
+            usage('The DNS client cookie provided had a length of %d, but only a length of %d is valid .' % (len(cookie), 8))
+            sys.exit(1)
+
+    return dns.edns.GenericOption(10, cookie)
+
+class CustomQueryMixin(object):
+    edns_options = []
+
 def usage(err=None):
     if err is not None:
         err += '\n\n'
     else:
         err = ''
-    sys.stderr.write('''%sUsage: dnsviz probe [options] [domain_name...]
+    sys.stderr.write('''%sUsage: %s %s [options] [domain_name...]
+
+Issue diagnostic DNS queries.
+
 Options:
-    -f <filename>  - read names from a file
-    -d <level>     - set debug level
-    -r <filename>  - read diagnostic queries from a file
-    -t <threads>   - specify number of threads to use for parallel queries
-    -4             - use IPv4 only
-    -6             - use IPv6 only
-    -b             - specify a source IPv4 or IPv6 address for queries
-    -u <url>       - URL for DNS looking glass
-    -k             - Do not verify TLS cert for DNS looking glass using HTTPS
-    -a <ancestor>  - query the ancestry of each domain name through ancestor
+    -f <filename>  - Read names from a file.
+    -d <level>     - Set debug level.
+    -r <filename>  - Read diagnostic queries from a file.
+    -t <threads>   - Use the specified number of threads for parallel queries.
+    -4             - Use IPv4 only.
+    -6             - Use IPv6 only.
+    -b <addr>      - Use the specified source IPv4 or IPv6 address for queries.
+    -u <url>       - Issue queries through the DNS looking glass at the
+                     specified URL.
+    -k             - Do not verify the TLS certificate for a DNS looking glass
+                     using HTTPS.
+    -a <ancestor>  - Query the ancestry of each domain name through the
+                     specified ancestor.
     -R <type>[,<type>...]
-                   - perform analysis using only the specified type(s)
+                   - Issue queries for only the specified type(s) during analysis.
     -s <server>[,<server>...]
-                   - designate servers for recursive analysis
-    -A             - query analysis against authoritative servers
+                   - Query the specified recursive server(s).
+    -A             - Query authoritative servers, instead of recursive servers.
     -x <domain>[+]:<server>[,<server>...]
-                   - designate authoritative servers explicitly for a domain
+                   - Query the specified authoritative servers for a domain.
     -N <domain>:<server>[,<server>...]
-                   - specify delegation information for a domain
+                   - Use the specified delegation information for a domain.
     -D <domain>:"<ds>"[,"<ds>"...]
-                   - specify DS records for a domain
-    -n             - use the NSID EDNS option
-    -e <subnet>[:<prefix>]
-                   - use the EDNS client subnet option with subnet/prefix
-    -E             - include EDNS compatibility diagnostics
-    -p             - make json output pretty instead of minimal
-    -o <filename>    - write the analysis to the specified file
-    -h             - display the usage and exit
-''' % (err))
+                   - Use the specified DS records for a domain.
+    -n             - Use the NSID EDNS option in queries.
+    -e <subnet>[:<prefix_len>]
+                   - Use the DNS client subnet option with the specified subnet
+                     and prefix length in queries.
+    -c <cookie>    - Use the specified DNS cookie value in queries.
+    -E             - Issue queries to check EDNS compatibility.
+    -o <filename>  - Write the analysis to the specified file.
+    -p             - Format JSON output with indentation and newlines.
+    -h             - Display the usage and exit.
+''' % (err, sys.argv[0], __name__.split('.')[-1]))
 
 def main(argv):
     global tm
+    global th_factories
     global resolver
     global bootstrap_resolver
     global explicit_delegations
@@ -720,7 +757,7 @@ def main(argv):
 
     try:
         try:
-            opts, args = getopt.getopt(argv[1:], 'f:d:l:c:r:t:64b:u:kmpo:a:R:x:N:D:ne:EAs:Fh')
+            opts, args = getopt.getopt(argv[1:], 'f:d:l:C:r:t:64b:u:kmpo:a:R:x:N:D:ne:c:EAs:Fh')
         except getopt.GetoptError as e:
             usage(str(e))
             sys.exit(1)
@@ -732,6 +769,7 @@ def main(argv):
         explicit_delegations = {}
         odd_ports = {}
         stop_at_explicit = {}
+        rdclass = dns.rdataclass.IN
         client_ipv4 = None
         client_ipv6 = None
         delegation_info = {}
@@ -959,7 +997,7 @@ def main(argv):
             if opts['-u'].startswith('https'):
                 vers0, vers1, vers2 = sys.version_info[:3]
                 if (2, 7, 9) > (vers0, vers1, vers2):
-                    sys.stderr.write('python version >= 2.7.9 is required to use a DNS looking glass with HTTPS.\n')
+                    logger.error('python version >= 2.7.9 is required to use a DNS looking glass with HTTPS.')
                     sys.exit(1)
 
             url = urlparse.urlparse(opts['-u'])
@@ -969,7 +1007,9 @@ def main(argv):
                 if url.hostname is not None:
                     usage('WebSocket URL must designate a local UNIX domain socket.')
                     sys.exit(1)
-                th_factories = (transport.DNSQueryTransportHandlerWebSocketFactory(url.path),)
+                th_factories = (transport.DNSQueryTransportHandlerWebSocketServerFactory(url.path),)
+            elif url.scheme == 'ssh':
+                th_factories = (transport.DNSQueryTransportHandlerRemoteCmdFactory(opts['-u']),)
             else:
                 usage('Unsupported URL scheme: "%s"' % opts['-u'])
                 sys.exit(1)
@@ -988,7 +1028,7 @@ def main(argv):
         # the following option is not documented in usage, as it doesn't
         # apply to most users
         try:
-            cache_level = int(opts['-c'])
+            cache_level = int(opts['-C'])
         except (KeyError, ValueError):
             cache_level = None
 
@@ -1018,9 +1058,6 @@ def main(argv):
             debug_level = logging.WARNING
         else:
             debug_level = logging.ERROR
-        handler = logging.StreamHandler()
-        handler.setLevel(debug_level)
-        logger.addHandler(handler)
         logger.setLevel(debug_level)
 
         if '-A' in opts:
@@ -1035,14 +1072,19 @@ def main(argv):
             else:
                 opt_r = opts['-r']
             try:
-                analysis_str = io.open(opt_r, 'r', encoding='utf-8').read()
+                with io.open(opt_r, 'r', encoding='utf-8') as fh:
+                    analysis_str = fh.read()
             except IOError as e:
                 logger.error('%s: "%s"' % (e.strerror, opts.get('-r', '-')))
                 sys.exit(3)
+            if not analysis_str:
+                if opt_r != sys.stdin.fileno():
+                    logger.error('No input.')
+                sys.exit(3)
             try:
                 analysis_structured = json.loads(analysis_str)
             except ValueError:
-                logger.error('There was an error parsing the json input: "%s"' % opts['-r'])
+                logger.error('There was an error parsing the JSON input: "%s"' % opts['-r'])
                 sys.exit(3)
 
             # check version
@@ -1061,7 +1103,7 @@ def main(argv):
                 logger.error('Version %d.%d of JSON input is incompatible with this software.' % (major_vers, minor_vers))
                 sys.exit(3)
 
-        names = []
+        names = OrderedDict()
         if '-f' in opts:
             if opts['-f'] == '-':
                 opts['-f'] = sys.stdin.fileno()
@@ -1079,7 +1121,8 @@ def main(argv):
                 except dns.exception.DNSException:
                     logger.error('The domain name was invalid: "%s"' % name)
                 else:
-                    names.append(name)
+                    if name not in names:
+                        names[name] = None
             f.close()
         else:
             if args:
@@ -1090,7 +1133,7 @@ def main(argv):
                 try:
                     args = analysis_structured['_meta._dnsviz.']['names']
                 except KeyError:
-                    logger.error('No names found in json input!')
+                    logger.error('No names found in JSON input!')
                     sys.exit(3)
             for name in args:
                 try:
@@ -1100,7 +1143,8 @@ def main(argv):
                 except dns.exception.DNSException:
                     logger.error('The domain name was invalid: "%s"' % name)
                 else:
-                    names.append(name)
+                    if name not in names:
+                        names[name] = None
 
         if '-p' in opts:
             kwargs = { 'indent': 4, 'separators': (',', ': ') }
@@ -1129,15 +1173,17 @@ def main(argv):
 
         flush = '-F' in opts
 
-        if '-n' in opts or '-e' in opts:
-            CustomQueryMixin.edns_options = []
-            if '-e' in opts:
-                CustomQueryMixin.edns_options.append(_get_ecs_option(opts['-e']))
-            if '-n' in opts:
-                CustomQueryMixin.edns_options.append(_get_nsid_option())
-            query_class_mixin = CustomQueryMixin
+        query_class_mixin = CustomQueryMixin
+        if '-e' in opts:
+            CustomQueryMixin.edns_options.append(_get_ecs_option(opts['-e']))
+        if '-n' in opts:
+            CustomQueryMixin.edns_options.append(_get_nsid_option())
+        if '-c' in opts:
+            if opts['-c']:
+                CustomQueryMixin.edns_options.append(_get_dns_cookie_option(opts['-c']))
         else:
-            query_class_mixin = None
+            # No cookie option was specified, so generate one
+            CustomQueryMixin.edns_options.append(_get_dns_cookie_option())
 
         name_objs = []
         if '-r' in opts:
@@ -1149,13 +1195,13 @@ def main(argv):
                 name_objs.append(OnlineDomainNameAnalysis.deserialize(name, analysis_structured, cache))
         else:
             if '-t' in opts:
-                a = cls(try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, rdtypes, explicit_only, dlv_domain, th_factories, processes)
+                a = cls(rdclass, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, rdtypes, explicit_only, dlv_domain, processes)
             else:
                 if cls.use_full_resolver:
                     _init_full_resolver()
                 else:
                     _init_stub_resolver()
-                a = cls(try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, rdtypes, explicit_only, dlv_domain, th_factories)
+                a = cls(rdclass, try_ipv4, try_ipv6, client_ipv4, client_ipv6, query_class_mixin, ceiling, edns_diagnostics, stop_at_explicit, cache_level, rdtypes, explicit_only, dlv_domain)
                 if flush:
                     fh.write('{')
                     a.analyze(names, _flush)
