@@ -6,7 +6,7 @@
 #
 # Copyright 2015-2016 VeriSign, Inc.
 #
-# Copyright 2016-2017 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -50,7 +50,7 @@ def usage(err=None):
         err += '\n\n'
     else:
         err = ''
-    sys.stderr.write('''%sUsage: dnsviz query [@global-server] [domain] [q-type] [q-class] {q-opt}
+    sys.stderr.write('''%sUsage: %s %s [@global-server] [domain] [q-type] [q-class] {q-opt}
            {global-d-opt} host [@local-server] {local-d-opt}
            [ host [@local-server] {local-d-opt} [...]]
 Where:  domain    is in the Domain Name System
@@ -70,7 +70,7 @@ Where:  domain    is in the Domain Name System
         global d-opts and servers (before host name) affect all queries.
         local d-opts and servers (after host name) affect only that lookup.
         -h                           (print help and exit)
-''' % (err))
+''' % (err, sys.argv[0], __name__.split('.')[-1]))
 
 class DVCommandLineQuery:
     def __init__(self, qname, rdtype, rdclass):
@@ -97,12 +97,12 @@ class DVCommandLineQuery:
                     if not arg:
                         raise ValueError()
                 except ValueError:
-                    usage('+trusted-key requires a filename argument.')
+                    sys.stderr.write('+trusted-key requires a filename argument.\n')
                     sys.exit(1)
                 else:
                     self.trusted_keys_file = arg
             else:
-                usage('Option "%s" not recognized.' % arg)
+                sys.stderr.write('Option "%s" not recognized.\n' % arg)
                 sys.exit(1)
 
     def process_nameservers(self, nameservers, use_ipv4, use_ipv6):
@@ -216,7 +216,7 @@ class DVCommandLine:
         try:
             if len(self.args[self.arg_index]) > 2:
                 if not has_arg:
-                    usage('"%s" option does not take arguments' % self.args[self.arg_index][:2])
+                    sys.stderr.write('"%s" option does not take arguments\n' % self.args[self.arg_index][:2])
                     sys.exit(1)
                 return self.args[self.arg_index][2:]
             else:
@@ -225,7 +225,7 @@ class DVCommandLine:
                 else:
                     self.arg_index += 1
                     if self.arg_index >= len(self.args):
-                        usage('"%s" option requires an argument' % self.args[self.arg_index - 1])
+                        sys.stderr.write('"%s" option requires an argument\n' % self.args[self.arg_index - 1])
                         sys.exit(1)
                     return self.args[self.arg_index]
         finally:
@@ -352,7 +352,7 @@ class DVCommandLine:
             self._get_arg(False)
             self.options['use_ipv4'] = True
         else:
-            usage('Option "%s" not recognized.' % self.args[self.arg_index][:2])
+            sys.stderr.write('Option "%s" not recognized.\n' % self.args[self.arg_index][:2])
             sys.exit(1)
 
     def _add_query_option(self, query):
