@@ -1,8 +1,6 @@
 #
 # This file is a part of DNSViz, a tool suite for DNS/DNSSEC monitoring,
-# analysis, and visualization.  This file (or some portion thereof) is a
-# derivative work authored by VeriSign, Inc., and created in 2014, based on
-# code originally developed at Sandia National Laboratories.
+# analysis, and visualization.
 # Created by Casey Deccio (casey@deccio.net)
 #
 # Copyright 2012-2014 Sandia Corporation. Under the terms of Contract
@@ -11,7 +9,7 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
-# Copyright 2016 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -31,28 +29,67 @@ from __future__ import unicode_literals
 
 import atexit
 import base64
+import binascii
+import logging
 import struct
 import hashlib
 import os
 import re
 
+from . import format as fmt
+lb2s = fmt.latin1_binary_to_string
+
+logger = logging.getLogger(__name__)
+
+ALG_TYPE_DNSSEC = 0
+ALG_TYPE_DIGEST = 1
+ALG_TYPE_NSEC3 = 2
+
+ALG_TYPE_DNSSEC_TEXT = [
+        'algorithm',
+        'digest algorithm',
+        'NSEC3 algorithm',
+]
+
+_crypto_sources = {
+        'M2Crypto >= 0.21.1': (set([1,5,7,8,10]), set([1,2,4]), set([1])),
+        'M2Crypto >= 0.24.0': (set([3,6,13,14]), set(), set()),
+        'M2Crypto >= 0.24.0 and either openssl < 1.1.0 or openssl >= 1.1.0 plus the OpenSSL GOST Engine': (set([12]), set([3]), set()),
+        'libnacl': (set([15]), set(), set()),
+}
+_logged_modules = set()
+
+_supported_algs = set()
+_supported_digest_algs = set()
+_supported_nsec3_algs = set([1])
 try:
     from M2Crypto import EVP, RSA
     from M2Crypto.m2 import hex_to_bn, bn_to_mpi
 except:
-    _supported_algs = set()
-    _supported_digest_algs = set()
+    pass
 else:
-    _supported_algs = set([1,5,7,8,10])
-    _supported_digest_algs = set([1,2,4])
+    _supported_algs.update(set([1,5,7,8,10]))
+    _supported_digest_algs.update(set([1,2,4]))
 
-_supported_nsec3_algs = set([1])
+try:
+    from libnacl.sign import Verifier as ed25519Verifier
+except ImportError:
+    pass
+else:
+    _supported_algs.add(15)
 
 GOST_PREFIX = b'\x30\x63\x30\x1c\x06\x06\x2a\x85\x03\x02\x02\x13\x30\x12\x06\x07\x2a\x85\x03\x02\x02\x23\x01\x06\x07\x2a\x85\x03\x02\x02\x1e\x01\x03\x43\x00\x04\x40'
+GOST_ENGINE_NAME = b'gost'
 GOST_DIGEST_NAME = b'GOST R 34.11-94'
 
+# python3/python2 dual compatibility
+if not isinstance(GOST_ENGINE_NAME, str):
+    GOST_ENGINE_NAME = lb2s(GOST_ENGINE_NAME)
+    GOST_DIGEST_NAME = lb2s(GOST_DIGEST_NAME)
+
 EC_NOCOMPRESSION = b'\x04'
 
+
 def _init_dynamic():
     try:
         Engine.load_dynamic()
@@ -96,9 +133,17 @@ def digest_alg_is_supported(alg):
 def nsec3_alg_is_supported(alg):
     return alg in _supported_nsec3_algs
 
+def _log_unsupported_alg(alg, alg_type):
+    for mod in _crypto_sources:
+        if alg in _crypto_sources[mod][alg_type]:
+            if mod not in _logged_modules:
+                _logged_modules.add(mod)
+                logger.warning('Warning: Without the installation of %s, cryptographic validation of DNSSEC %s %d (and possibly others) is not supported.' % (mod, ALG_TYPE_DNSSEC_TEXT[alg_type], alg))
+            return
+
 def _gost_init():
     try:
-        gost = Engine.Engine(b'gost')
+        gost = Engine.Engine(GOST_ENGINE_NAME)
         gost.init()
         gost.set_default()
     except ValueError:
@@ -107,7 +152,7 @@ def _gost_init():
 def _gost_cleanup():
     from M2Crypto import Engine
     try:
-        gost = Engine.Engine(b'gost')
+        gost = Engine.Engine(GOST_ENGINE_NAME)
     except ValueError:
         pass
     else:
@@ -137,6 +182,7 @@ else:
 
 def validate_ds_digest(digest_alg, digest, dnskey_msg):
     if not digest_alg_is_supported(digest_alg):
+        _log_unsupported_alg(digest_alg, ALG_TYPE_DIGEST)
         return None
 
     if digest_alg == 1:
@@ -162,39 +208,30 @@ def validate_ds_digest(digest_alg, digest, dnskey_msg):
 
 def _dnskey_to_dsa(key):
     # get T
-    t, = struct.unpack(b'B',key[0])
+    t = key[0]
+    # python3/python2 dual compatibility
+    if not isinstance(t, int):
+        t = ord(t)
     offset = 1
 
     # get Q
     new_offset = offset+20
-    q = b''
-    for c in key[offset:new_offset]:
-        q += b'%02x' % struct.unpack(b'B',c)[0]
-    q = bn_to_mpi(hex_to_bn(q))
+    q = bn_to_mpi(hex_to_bn(binascii.hexlify(key[offset:new_offset])))
     offset = new_offset
 
     # get P
     new_offset = offset+64+(t<<3)
-    p = b''
-    for c in key[offset:new_offset]:
-        p += b'%02x' % struct.unpack(b'B',c)[0]
-    p = bn_to_mpi(hex_to_bn(p))
+    p = bn_to_mpi(hex_to_bn(binascii.hexlify(key[offset:new_offset])))
     offset = new_offset
 
     # get G
     new_offset = offset+64+(t<<3)
-    g = b''
-    for c in key[offset:new_offset]:
-        g += b'%02x' % struct.unpack(b'B',c)[0]
-    g = bn_to_mpi(hex_to_bn(g))
+    g = bn_to_mpi(hex_to_bn(binascii.hexlify(key[offset:new_offset])))
     offset = new_offset
 
     # get Y
     new_offset = offset+64+(t<<3)
-    y = b''
-    for c in key[offset:new_offset]:
-        y += b'%02x' % struct.unpack(b'B',c)[0]
-    y = bn_to_mpi(hex_to_bn(y))
+    y = bn_to_mpi(hex_to_bn(binascii.hexlify(key[offset:new_offset])))
     offset = new_offset
 
     # create the DSA public key
@@ -203,9 +240,12 @@ def _dnskey_to_dsa(key):
 def _dnskey_to_rsa(key):
     try:
         # get the exponent length
-        e_len, = struct.unpack(b'B',key[0])
+        e_len = key[0]
     except IndexError:
         return None
+    # python3/python2 dual compatibility
+    if not isinstance(e_len, int):
+        e_len = ord(e_len)
 
     offset = 1
     if e_len == 0:
@@ -213,17 +253,11 @@ def _dnskey_to_rsa(key):
         offset = 3
 
     # get the exponent
-    e = b''
-    for c in key[offset:offset+e_len]:
-        e += b'%02x' % struct.unpack(b'B',c)[0]
-    e = bn_to_mpi(hex_to_bn(e))
+    e = bn_to_mpi(hex_to_bn(binascii.hexlify(key[offset:offset+e_len])))
     offset += e_len
 
     # get the modulus
-    n = b''
-    for c in key[offset:]:
-        n += b'%02x' % struct.unpack(b'B',c)[0]
-    n = bn_to_mpi(hex_to_bn(n))
+    n = bn_to_mpi(hex_to_bn(binascii.hexlify(key[offset:])))
 
     # create the RSA public key
     rsa = RSA.new_pub_key((e,n))
@@ -234,7 +268,7 @@ def _dnskey_to_rsa(key):
 
 def _dnskey_to_gost(key):
     der = GOST_PREFIX + key
-    pem = bytes('-----BEGIN PUBLIC KEY-----\n'+base64.encodestring(der)+'-----END PUBLIC KEY-----')
+    pem = b'-----BEGIN PUBLIC KEY-----\n'+base64.encodestring(der)+b'-----END PUBLIC KEY-----'
 
     return EVP.load_key_string_pubkey(pem)
 
@@ -276,23 +310,20 @@ def _validate_rrsig_dsa(alg, sig, msg, key):
     pubkey = _dnskey_to_dsa(key)
 
     # get T
-    t, = struct.unpack(b'B',sig[0])
+    t = sig[0]
+    # python3/python2 dual compatibility
+    if not isinstance(t, int):
+        t = ord(t)
     offset = 1
 
     # get R
     new_offset = offset+20
-    r = b''
-    for c in sig[offset:new_offset]:
-        r += b'%02x' % struct.unpack(b'B',c)[0]
-    r = bn_to_mpi(hex_to_bn(r))
+    r = bn_to_mpi(hex_to_bn(binascii.hexlify(sig[offset:new_offset])))
     offset = new_offset
 
     # get S
     new_offset = offset+20
-    s = b''
-    for c in sig[offset:new_offset]:
-        s += b'%02x' % struct.unpack(b'B',c)[0]
-    s = bn_to_mpi(hex_to_bn(s))
+    s = bn_to_mpi(hex_to_bn(binascii.hexlify(sig[offset:new_offset])))
     offset = new_offset
 
     md = EVP.MessageDigest('sha1')
@@ -339,18 +370,12 @@ def _validate_rrsig_ec(alg, sig, msg, key):
 
     # get R
     new_offset = offset+sigsize//2
-    r = b''
-    for c in sig[offset:new_offset]:
-        r += b'%02x' % struct.unpack(b'B',c)[0]
-    r = bn_to_mpi(hex_to_bn(r))
+    r = bn_to_mpi(hex_to_bn(binascii.hexlify(sig[offset:new_offset])))
     offset = new_offset
 
     # get S
     new_offset = offset+sigsize//2
-    s = b''
-    for c in sig[offset:new_offset]:
-        s += b'%02x' % struct.unpack(b'B',c)[0]
-    s = bn_to_mpi(hex_to_bn(s))
+    s = bn_to_mpi(hex_to_bn(binascii.hexlify(sig[offset:new_offset])))
     offset = new_offset
 
     md = EVP.MessageDigest(alg)
@@ -359,8 +384,16 @@ def _validate_rrsig_ec(alg, sig, msg, key):
 
     return pubkey.verify_dsa(digest, r, s) == 1
 
+def _validate_rrsig_ed25519(alg, sig, msg, key):
+    try:
+        verifier = ed25519Verifier(binascii.hexlify(key))
+        return verifier.verify(sig + msg) == msg
+    except ValueError:
+        return False
+
 def validate_rrsig(alg, sig, msg, key):
     if not alg_is_supported(alg):
+        _log_unsupported_alg(alg, ALG_TYPE_DNSSEC)
         return None
 
     # create an RSA key object for RSA keys
@@ -372,9 +405,12 @@ def validate_rrsig(alg, sig, msg, key):
         return _validate_rrsig_gost(alg, sig, msg, key)
     elif alg in (13,14):
         return _validate_rrsig_ec(alg, sig, msg, key)
+    elif alg in (15,):
+        return _validate_rrsig_ed25519(alg, sig, msg, key)
 
 def get_digest_for_nsec3(val, salt, alg, iterations):
     if not nsec3_alg_is_supported(alg):
+        _log_unsupported_alg(alg, ALG_TYPE_NSEC3)
         return None
 
     if alg == 1:
