@@ -1,8 +1,6 @@
 #
 # This file is a part of DNSViz, a tool suite for DNS/DNSSEC monitoring,
-# analysis, and visualization.  This file (or some portion thereof) is a
-# derivative work authored by VeriSign, Inc., and created in 2014, based on
-# code originally developed at Sandia National Laboratories.
+# analysis, and visualization.
 # Created by Casey Deccio (casey@deccio.net)
 #
 # Copyright 2012-2014 Sandia Corporation. Under the terms of Contract
@@ -11,7 +9,7 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
-# Copyright 2016 Casey Deccio.
+# Copyright 2016-2019 Casey Deccio
 #
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -40,7 +38,8 @@ import dns.name, dns.rdatatype
 DNSKEY_FLAGS = {'ZONE': 0x0100, 'SEP': 0x0001, 'revoke': 0x0080}
 DNSKEY_PROTOCOLS = { 3: 'DNSSEC' }
 DNSKEY_ALGORITHMS = { 1: 'RSA/MD5', 2: 'Diffie-Hellman', 3: 'DSA/SHA1', 5: 'RSA/SHA-1', 6: 'DSA-NSEC3-SHA1', 7: 'RSASHA1-NSEC3-SHA1', \
-        8: 'RSA/SHA-256', 10: 'RSA/SHA-512', 12: 'GOST R 34.10-2001', 13: 'ECDSA Curve P-256 with SHA-256', 14: 'ECDSA Curve P-384 with SHA-384' }
+        8: 'RSA/SHA-256', 10: 'RSA/SHA-512', 12: 'GOST R 34.10-2001', 13: 'ECDSA Curve P-256 with SHA-256', 14: 'ECDSA Curve P-384 with SHA-384',
+        15: 'Ed25519', 16: 'Ed448' }
 DS_DIGEST_TYPES = { 1: 'SHA-1', 2: 'SHA-256', 3: 'GOST 34.11-94', 4: 'SHA-384' }
 
 NSEC3_FLAGS = {'OPTOUT': 0x01}
@@ -157,14 +156,16 @@ def format_nsec3_name(name):
 def format_nsec3_rrset_text(nsec3_rrset_text):
     return re.sub(r'^(\d+\s+\d+\s+\d+\s+\S+\s+)([0-9a-zA-Z]+)', lambda x: '%s%s' % (x.group(1), x.group(2).upper()), nsec3_rrset_text).rstrip('.')
 
-def humanize_name(name, idn=False):
+def humanize_name(name, idn=False, canonicalize=True):
+    if canonicalize:
+        name = name.canonicalize()
     if idn:
         try:
-            name = name.canonicalize().to_unicode()
+            name = name.to_unicode()
         except UnicodeError:
-            name = lb2s(name.canonicalize().to_text())
+            name = lb2s(name.to_text())
     else:
-        name = lb2s(name.canonicalize().to_text())
+        name = lb2s(name.to_text())
     if name == '.':
         return name
     return name.rstrip('.')
