@@ -5,6 +5,8 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
+# Copyright 2016-2019 Casey Deccio
+#
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
@@ -294,7 +296,7 @@ class FullResolver:
 
     default_th_factory = transport.DNSQueryTransportHandlerDNSFactory()
 
-    def __init__(self, hints=util.get_root_hints(), query_cls=(query.QuickDNSSECQuery, query.RobustDNSSECQuery), client_ipv4=None, client_ipv6=None, odd_ports=None, transport_manager=None, th_factories=None, max_ttl=None):
+    def __init__(self, hints=util.get_root_hints(), query_cls=(query.QuickDNSSECQuery, query.DiagnosticQuery), client_ipv4=None, client_ipv6=None, odd_ports=None, cookie_standin=None, transport_manager=None, th_factories=None, max_ttl=None):
 
         self._hints = hints
         self._query_cls = query_cls
@@ -313,6 +315,8 @@ class FullResolver:
 
         self._max_ttl = max_ttl
 
+        self._cookie_standin = cookie_standin
+        self._cookie_jar = {}
         self._cache = {}
         self._expirations = []
         self._cache_lock = threading.Lock()
@@ -445,13 +449,7 @@ class FullResolver:
             responses[query_tuple] = self.query(query_tuple[0], query_tuple[1], query_tuple[2])
         return responses
 
-    def _query(self, qname, rdtype, rdclass, level, max_source, starting_domain=None):
-        self.expire_cache()
-
-        # check for max chain length
-        if level > self.MAX_CHAIN:
-            raise ServFail('SERVFAIL - resolution chain too long')
-
+    def _get_answer(self, qname, rdtype, rdclass, max_source):
         # first check cache for answer
         entry = self.cache_get(qname, rdtype)
         if entry is not None and entry.source <= max_source:
@@ -461,10 +459,23 @@ class FullResolver:
         if self.SRC_ADDITIONAL <= max_source and (qname, rdtype) in self._hints:
             return [self._hints[(qname, rdtype)], dns.rcode.NOERROR]
 
+        return None
+
+    def _query(self, qname, rdtype, rdclass, level, max_source, starting_domain=None):
+        self.expire_cache()
+
+        # check for max chain length
+        if level > self.MAX_CHAIN:
+            raise ServFail('SERVFAIL - resolution chain too long')
+
+        ans = self._get_answer(qname, rdtype, rdclass, max_source)
+        if ans:
+            return ans
+
         # next check cache for alias
-        entry = self.cache_get(qname, dns.rdatatype.CNAME)
-        if entry is not None and entry.rrset is not None:
-            return [entry.rrset] + self._query(entry.rrset[0].target, rdtype, rdclass, level + 1, max_source)
+        ans = self._get_answer(qname, dns.rdatatype.CNAME, rdclass, max_source)
+        if ans:
+            return [ans[0]] + self._query(entry.rrset[0].target, rdtype, rdclass, level + 1, max_source)
 
         # now check for closest enclosing NS, DNAME, or hint
         closest_zone = qname
@@ -488,22 +499,20 @@ class FullResolver:
                     return [entry.rrset, cname_rrset] + self._query(cname_rrset[0].target, rdtype, rdclass, level + 1, max_source)
 
             # look for NS records in cache
-            entry = self.cache_get(closest_zone, dns.rdatatype.NS)
-            if entry is not None:
-                if entry.rrset is not None:
-                    ns_rrset = entry.rrset
-                    for rdata in entry.rrset:
-                        ns_names[rdata.target] = None
-
-            # look for NS records in hints
-            else:
-                try:
-                    ns_rrset = self._hints[(closest_zone, dns.rdatatype.NS)]
-                except KeyError:
-                    pass
-                else:
-                    for rdata in ns_rrset:
-                        ns_names[rdata.target] = None
+            ans = self._get_answer(closest_zone, dns.rdatatype.NS, rdclass, self.SRC_ADDITIONAL)
+            if ans and ans[0] is not None:
+                ns_rrset = ans[0]
+                for ns_rdata in ans[0]:
+                    addrs = set()
+                    for a_rdtype in dns.rdatatype.A, dns.rdatatype.AAAA:
+                        ans1 = self._get_answer(ns_rdata.target, a_rdtype, rdclass, self.SRC_ADDITIONAL)
+                        if ans1 and ans1[0]:
+                            for a_rdata in ans1[0]:
+                                addrs.add(IPAddr(a_rdata.address))
+                    if addrs:
+                        ns_names[ns_rdata.target] = addrs
+                    else:
+                        ns_names[ns_rdata.target] = None
 
             # if there were NS records associated with the names, then
             # no need to continue
@@ -531,6 +540,7 @@ class FullResolver:
             ns_names_without_addresses = list(set(ns_names).difference(ns_names_with_addresses))
             random.shuffle(ns_names_without_addresses)
             all_ns_names = ns_names_with_addresses + ns_names_without_addresses
+            previous_valid_answer = set()
 
             for query_cls in self._query_cls:
                 # query each server until we get a match
@@ -555,12 +565,12 @@ class FullResolver:
                                 for rdata in a_rrset:
                                     ns_names[ns_name].add(IPAddr(rdata.address))
 
-                    for server in ns_names[ns_name]:
+                    for server in ns_names[ns_name].difference(previous_valid_answer):
                         # server disallowed by policy
                         if not self._allow_server(server):
                             continue
 
-                        q = query_cls(qname, rdtype, rdclass, (server,), bailiwick, self._client_ipv4, self._client_ipv6, self._odd_ports.get((bailiwick, server), 53))
+                        q = query_cls(qname, rdtype, rdclass, (server,), bailiwick, self._client_ipv4, self._client_ipv6, self._odd_ports.get((bailiwick, server), 53), cookie_jar=self._cookie_jar, cookie_standin=self._cookie_standin)
                         q.execute(tm=self._transport_manager, th_factories=self._th_factories)
                         is_referral = False
 
@@ -571,46 +581,94 @@ class FullResolver:
                         server1, client_response = list(q.responses.items())[0]
                         client, response = list(client_response.items())[0]
 
-                        if response.is_valid_response() and response.is_complete_response():
-                            soa_rrset = None
-                            rcode = response.message.rcode()
+                        server_cookie = response.get_server_cookie()
+                        if server_cookie is not None:
+                            self._cookie_jar[server1] = server_cookie
 
-                            # response is acceptable
+                        if not (response.is_valid_response() and response.is_complete_response()):
+                            continue
+
+                        previous_valid_answer.add(server)
+
+                        soa_rrset = None
+                        rcode = response.message.rcode()
+
+                        # response is acceptable
+                        try:
+                            # first check for exact match
+                            ret = [[x for x in response.message.answer if x.name == qname and x.rdtype == rdtype and x.rdclass == rdclass][0]]
+                        except IndexError:
                             try:
-                                # first check for exact match
-                                ret = [[x for x in response.message.answer if x.name == qname and x.rdtype == rdtype and x.rdclass == rdclass][0]]
+                                # now look for DNAME
+                                dname_rrset = [x for x in response.message.answer if qname.is_subdomain(x.name) and qname != x.name and x.rdtype == dns.rdatatype.DNAME and x.rdclass == rdclass][0]
                             except IndexError:
                                 try:
-                                    # now look for DNAME
-                                    dname_rrset = [x for x in response.message.answer if qname.is_subdomain(x.name) and qname != x.name and x.rdtype == dns.rdatatype.DNAME and x.rdclass == rdclass][0]
+                                    # now look for CNAME
+                                    cname_rrset = [x for x in response.message.answer if x.name == qname and x.rdtype == dns.rdatatype.CNAME and x.rdclass == rdclass][0]
                                 except IndexError:
+                                    ret = [None]
+                                    # no answer
                                     try:
-                                        # now look for CNAME
-                                        cname_rrset = [x for x in response.message.answer if x.name == qname and x.rdtype == dns.rdatatype.CNAME and x.rdclass == rdclass][0]
+                                        soa_rrset = [x for x in response.message.authority if qname.is_subdomain(x.name) and x.rdtype == dns.rdatatype.SOA][0]
                                     except IndexError:
-                                        ret = [None]
-                                        # no answer
-                                        try:
-                                            soa_rrset = [x for x in response.message.authority if qname.is_subdomain(x.name) and x.rdtype == dns.rdatatype.SOA][0]
-                                        except IndexError:
-                                            pass
-                                    # cache the NS RRset
-                                    else:
-                                        cname_rrset = [x for x in response.message.answer if x.name == qname and x.rdtype == dns.rdatatype.CNAME and x.rdclass == rdclass][0]
-                                        ret = [cname_rrset]
+                                        pass
+                                # cache the NS RRset
                                 else:
-                                    # handle DNAME: return the DNAME, CNAME and (recursively) its chain
-                                    cname_rrset = Response.cname_from_dname(qname, dname_rrset)
-                                    ret = [dname_rrset, cname_rrset]
+                                    cname_rrset = [x for x in response.message.answer if x.name == qname and x.rdtype == dns.rdatatype.CNAME and x.rdclass == rdclass][0]
+                                    ret = [cname_rrset]
+                            else:
+                                # handle DNAME: return the DNAME, CNAME and (recursively) its chain
+                                cname_rrset = Response.cname_from_dname(qname, dname_rrset)
+                                ret = [dname_rrset, cname_rrset]
 
-                            if response.is_referral(qname, rdtype, bailiwick):
-                                is_referral = True
-                                a_rrsets = {}
-                                min_ttl = None
-                                ret = None
+                        if response.is_referral(qname, rdtype, rdclass, bailiwick):
+                            is_referral = True
+                            a_rrsets = {}
+                            min_ttl = None
+                            ret = None
+
+                            # if response is referral, then we follow it
+                            ns_rrset = [x for x in response.message.authority if qname.is_subdomain(x.name) and x.rdtype == dns.rdatatype.NS][0]
+                            ns_names = response.ns_ip_mapping_from_additional(ns_rrset.name, bailiwick)
+                            for ns_name in ns_names:
+                                if not ns_names[ns_name]:
+                                    ns_names[ns_name] = None
+                                else: # name is in bailiwick
+                                    for a_rdtype in (dns.rdatatype.A, dns.rdatatype.AAAA):
+                                        try:
+                                            a_rrsets[a_rdtype] = response.message.find_rrset(response.message.additional, ns_name, a_rdtype, dns.rdataclass.IN)
+                                        except KeyError:
+                                            pass
+                                        else:
+                                            if min_ttl is None or a_rrsets[a_rdtype].ttl < min_ttl:
+                                                min_ttl = a_rrsets[a_rdtype].ttl
+
+                                    for a_rdtype in (dns.rdatatype.A, dns.rdatatype.AAAA):
+                                        if a_rdtype in a_rrsets:
+                                            a_rrsets[a_rdtype].update_ttl(min_ttl)
+                                            self.cache_put(ns_name, a_rdtype, a_rrsets[a_rdtype], self.SRC_ADDITIONAL, dns.rcode.NOERROR, None, None)
+                                        else:
+                                            self.cache_put(ns_name, a_rdtype, None, self.SRC_ADDITIONAL, dns.rcode.NOERROR, None, min_ttl)
+
+                            if min_ttl is not None:
+                                ns_rrset.update_ttl(min_ttl)
+
+                            # cache the NS RRset
+                            self.cache_put(ns_rrset.name, dns.rdatatype.NS, ns_rrset, self.SRC_NONAUTH_AUTH, rcode, None, None)
+                            break
+
+                        elif response.is_authoritative():
+                            terminal = True
+                            a_rrsets = {}
+                            min_ttl = None
+
+                            # if response is authoritative (and not a referral), then we return it
+                            try:
+                                ns_rrset = [x for x in  response.message.answer + response.message.authority if qname.is_subdomain(x.name) and x.rdtype == dns.rdatatype.NS][0]
+                            except IndexError:
+                                pass
+                            else:
 
-                                # if response is referral, then we follow it
-                                ns_rrset = [x for x in response.message.authority if qname.is_subdomain(x.name) and x.rdtype == dns.rdatatype.NS][0]
                                 ns_names = response.ns_ip_mapping_from_additional(ns_rrset.name, bailiwick)
                                 for ns_name in ns_names:
                                     if not ns_names[ns_name]:
@@ -635,62 +693,22 @@ class FullResolver:
                                 if min_ttl is not None:
                                     ns_rrset.update_ttl(min_ttl)
 
-                                # cache the NS RRset
-                                self.cache_put(ns_rrset.name, dns.rdatatype.NS, ns_rrset, self.SRC_NONAUTH_AUTH, rcode, None, None)
-                                break
+                                self.cache_put(ns_rrset.name, dns.rdatatype.NS, ns_rrset, self.SRC_AUTH_AUTH, rcode, None, None)
 
-                            elif response.is_authoritative():
-                                terminal = True
-                                a_rrsets = {}
-                                min_ttl = None
+                            if ret[-1] == None:
+                                self.cache_put(qname, rdtype, None, self.SRC_AUTH_ANS, rcode, soa_rrset, None)
 
-                                # if response is authoritative (and not a referral), then we return it
-                                try:
-                                    ns_rrset = [x for x in  response.message.answer + response.message.authority if qname.is_subdomain(x.name) and x.rdtype == dns.rdatatype.NS][0]
-                                except IndexError:
-                                    pass
-                                else:
+                            else:
+                                for rrset in ret:
+                                    self.cache_put(rrset.name, rrset.rdtype, rrset, self.SRC_AUTH_ANS, rcode, None, None)
 
-                                    ns_names = response.ns_ip_mapping_from_additional(ns_rrset.name, bailiwick)
-                                    for ns_name in ns_names:
-                                        if not ns_names[ns_name]:
-                                            ns_names[ns_name] = None
-                                        else: # name is in bailiwick
-                                            for a_rdtype in (dns.rdatatype.A, dns.rdatatype.AAAA):
-                                                try:
-                                                    a_rrsets[a_rdtype] = response.message.find_rrset(response.message.additional, ns_name, a_rdtype, dns.rdataclass.IN)
-                                                except KeyError:
-                                                    pass
-                                                else:
-                                                    if min_ttl is None or a_rrsets[a_rdtype].ttl < min_ttl:
-                                                        min_ttl = a_rrsets[a_rdtype].ttl
+                                if ret[-1].rdtype == dns.rdatatype.CNAME:
+                                    ret += self._query(ret[-1][0].target, rdtype, rdclass, level + 1, self.SRC_NONAUTH_ANS)
+                                    terminal = False
 
-                                            for a_rdtype in (dns.rdatatype.A, dns.rdatatype.AAAA):
-                                                if a_rdtype in a_rrsets:
-                                                    a_rrsets[a_rdtype].update_ttl(min_ttl)
-                                                    self.cache_put(ns_name, a_rdtype, a_rrsets[a_rdtype], self.SRC_ADDITIONAL, dns.rcode.NOERROR, None, None)
-                                                else:
-                                                    self.cache_put(ns_name, a_rdtype, None, self.SRC_ADDITIONAL, dns.rcode.NOERROR, None, min_ttl)
-
-                                    if min_ttl is not None:
-                                        ns_rrset.update_ttl(min_ttl)
-
-                                    self.cache_put(ns_rrset.name, dns.rdatatype.NS, ns_rrset, self.SRC_AUTH_AUTH, rcode, None, None)
-
-                                if ret[-1] == None:
-                                    self.cache_put(qname, rdtype, None, self.SRC_AUTH_ANS, rcode, soa_rrset, None)
-
-                                else:
-                                    for rrset in ret:
-                                        self.cache_put(rrset.name, rrset.rdtype, rrset, self.SRC_AUTH_ANS, rcode, None, None)
-
-                                    if ret[-1].rdtype == dns.rdatatype.CNAME:
-                                        ret += self._query(ret[-1][0].target, rdtype, rdclass, level + 1, self.SRC_NONAUTH_ANS)
-                                        terminal = False
-
-                                if terminal:
-                                    ret.append(rcode)
-                                return ret
+                            if terminal:
+                                ret.append(rcode)
+                            return ret
 
                     # if referral, then break
                     if is_referral:
