@@ -1,8 +1,6 @@
 #
 # This file is a part of DNSViz, a tool suite for DNS/DNSSEC monitoring,
-# analysis, and visualization.  This file (or some portion thereof) is a
-# derivative work authored by VeriSign, Inc., and created in 2014, based on
-# code originally developed at Sandia National Laboratories.
+# analysis, and visualization.
 # Created by Casey Deccio (casey@deccio.net)
 #
 # Copyright 2012-2014 Sandia Corporation. Under the terms of Contract
@@ -11,6 +9,8 @@
 #
 # Copyright 2014-2016 VeriSign, Inc.
 #
+# Copyright 2016-2019 Casey Deccio
+#
 # DNSViz is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
@@ -28,8 +28,9 @@
 from __future__ import unicode_literals
 
 import base64
+import binascii
+import copy
 import errno
-import cgi
 import codecs
 import datetime
 import hashlib
@@ -45,6 +46,12 @@ try:
 except ImportError:
     from ordereddict import OrderedDict
 
+# python3/python2 dual compatibility
+try:
+    from html import escape
+except ImportError:
+    from cgi import escape
+
 import dns.flags, dns.message, dns.rcode, dns.rdataclass, dns.rdatatype, dns.rrset
 
 from . import base32
@@ -57,7 +64,7 @@ lb2s = fmt.latin1_binary_to_string
 class DNSResponse:
     '''A DNS response, including meta information'''
 
-    def __init__(self, message, msg_size, error, errno1, history, response_time, query, review_history=True):
+    def __init__(self, message, msg_size, error, errno1, history, response_time, query, server_cookie, server_cookie_status, review_history=True):
         self.message = message
         self.msg_size = msg_size
         self.error = error
@@ -66,6 +73,8 @@ class DNSResponse:
         self.response_time = response_time
 
         self.query = query
+        self.server_cookie = server_cookie
+        self.server_cookie_status = server_cookie_status
 
         self.effective_flags = None
         self.effective_edns = None
@@ -73,6 +82,7 @@ class DNSResponse:
         self.effective_edns_flags = None
         self.effective_edns_options = None
         self.effective_tcp = None
+        self.effective_server_cookie_status = None
 
         self.udp_attempted = None
         self.udp_responsive = None
@@ -93,36 +103,154 @@ class DNSResponse:
     def __repr__(self):
         return '<%s: "%s">' % (self.__class__.__name__, str(self))
 
-    def initial_query_tag(self):
-        s = ''
-        if self.query.tcp:
-            s += 'TCP_'
+    @classmethod
+    def _query_tag_bind(cls, tcp, flags, edns, edns_flags, edns_max_udp_payload, edns_options, qname):
+        s = []
+        if flags & dns.flags.RD:
+            s.append('+')
         else:
-            s += 'UDP_'
-        s += '%d_' % self.query.flags
-        if self.query.edns < 0:
-            s += 'NOEDNS'
-        else:
-            s += 'EDNS%d_%d_%d' % (self.query.edns, (self.query.edns_flags & 0xffff), self.query.edns_max_udp_payload)
-            for opt in self.query.edns_options:
-                s += '_%d' % opt.otype
+            s.append('-')
+        if edns >= 0:
+            s.append('E(%d)' % (edns))
+        if tcp:
+            s.append('T')
+        if edns >= 0 and edns_flags & dns.flags.DO:
+            s.append('D')
+        if flags & dns.flags.CD:
+            s.append('C')
+        # Flags other than the ones commonly seen in queries
+        if flags & dns.flags.AD:
+            s.append('A')
+        if flags & dns.flags.AA:
+            s.append('a')
+        if flags & dns.flags.TC:
+            s.append('t')
+        if flags & dns.flags.RA:
+            s.append('r')
+        if edns >= 0:
+            # EDNS max UDP payload
+            s.append('P(%d)' % edns_max_udp_payload)
+            # EDNS flags other than DO
+            if edns_flags & ~dns.flags.DO:
+                s.append('F(0x%x)' % edns_flags)
+            # other options
+            for opt in edns_options:
+                if opt.otype == 3:
+                    # NSID
+                    s.append('N')
+                elif opt.otype == 8:
+                    # EDNS Client Subnet
+                    s.append('s')
+                elif opt.otype == 10:
+                    # DNS cookies
+                    s.append('K')
+        if qname.to_text() != qname.to_text().lower():
+            s.append('X')
         return s
 
-    def effective_query_tag(self):
+    @classmethod
+    def _query_tag_human(cls, tcp, flags, edns, edns_flags, edns_max_udp_payload, edns_options, qname):
         s = ''
-        if self.effective_tcp:
+        if tcp:
             s += 'TCP_'
         else:
             s += 'UDP_'
-        s += '%d_' % self.effective_flags
-        if self.effective_edns < 0:
-            s += 'NOEDNS'
+
+        if flags & dns.flags.RD:
+            s += '+'
         else:
-            s += 'EDNS%d_%d_%d' % (self.effective_edns, (self.effective_edns_flags & 0xffff), self.effective_edns_max_udp_payload)
-            for opt in self.effective_edns_options:
-                s += '_%d' % opt.otype
+            s += '-'
+        if flags & dns.flags.CD:
+            s += 'C'
+        # Flags other than the ones commonly seen in queries
+        if flags & dns.flags.AD:
+            s += 'A'
+        if flags & dns.flags.AA:
+            s += 'a'
+        if flags & dns.flags.TC:
+            s += 't'
+        if flags & dns.flags.RA:
+            s += 'r'
+        s += '_'
+
+        if edns < 0:
+            s += 'NOEDNS_'
+        else:
+            s += 'EDNS%d_' % (edns)
+
+            # EDNS max UDP payload
+            s += '%d_' % edns_max_udp_payload
+
+            if edns_flags & dns.flags.DO:
+                s += 'D'
+            # EDNS flags other than DO
+            if edns_flags & ~dns.flags.DO:
+                s += '%d' % edns_flags
+            s += '_'
+
+            # other options
+            for opt in edns_options:
+                if opt.otype == 3:
+                    # NSID
+                    s += 'N'
+                elif opt.otype == 8:
+                    # EDNS Client Subnet
+                    s += 's'
+                elif opt.otype == 10:
+                    # DNS cookies
+                    s += 'K'
+                else:
+                    # DNS cookies
+                    s += 'O(%d)' % opt.otype
+
+        if qname.to_text() != qname.to_text().lower():
+            s += '_0x20'
         return s
 
+    def request_cookie_tag(self):
+        from . import query as Q
+
+        if self.effective_server_cookie_status == Q.DNS_COOKIE_NO_COOKIE:
+            return 'NO_COOKIE'
+        elif self.effective_server_cookie_status == Q.DNS_COOKIE_IMPROPER_LENGTH:
+            return 'MALFORMED_COOKIE'
+        elif self.effective_server_cookie_status == Q.DNS_COOKIE_CLIENT_COOKIE_ONLY:
+            return 'CLIENT_COOKIE_ONLY'
+        elif self.effective_server_cookie_status == Q.DNS_COOKIE_SERVER_COOKIE_FRESH:
+            return 'VALID_SERVER_COOKIE'
+        elif self.effective_server_cookie_status == Q.DNS_COOKIE_SERVER_COOKIE_BAD:
+            return 'INVALID_SERVER_COOKIE'
+        else:
+            raise Exception('Unknown cookie status!')
+
+    def response_cookie_tag(self):
+
+        if self.message is None:
+            return 'ERROR'
+
+        if self.message.edns < 0:
+            return 'NO_EDNS'
+
+        try:
+            cookie_opt = [o for o in self.message.options if o.otype == 10][0]
+        except IndexError:
+            return 'NO_COOKIE_OPT'
+
+        if len(cookie_opt.data) < 8 or len(cookie_opt.data) > 40:
+            return 'MALFORMED_COOKIE'
+
+        elif len(cookie_opt.data) == 8:
+            return 'CLIENT_COOKIE_ONLY'
+
+        else:
+            return 'CLIENT_AND_SERVER_COOKIE'
+
+    def initial_query_tag(self):
+        return ''.join(self._query_tag_human(self.query.tcp, self.query.flags, self.query.edns, self.query.edns_flags, self.query.edns_max_udp_payload, self.query.edns_options, self.query.qname))
+
+    def effective_query_tag(self):
+        return ''.join(self._query_tag_human(self.effective_tcp, self.effective_flags, self.effective_edns, self.effective_edns_flags, self.query.edns_max_udp_payload, self.effective_edns_options, self.query.qname))
+
     def section_rr_count(self, section):
         if self.message is None:
             return None
@@ -152,19 +280,34 @@ class DNSResponse:
             t += retry.response_time
         return t
 
+    def get_cookie_opt(self):
+        if self.message is None:
+            return None
+        try:
+            return [o for o in self.message.options if o.otype == 10][0]
+        except IndexError:
+            return None
+
+    def get_server_cookie(self):
+        cookie_opt = self.get_cookie_opt()
+        if cookie_opt is not None and len(cookie_opt.data) > 8:
+            return cookie_opt.data[8:]
+        return None
+
     def copy(self):
-        clone = DNSResponse(self.message, self.msg_size, self.error, self.errno, self.history, self.response_time, self.query, review_history=False)
-        clone.set_effective_request_options(self.effective_flags, self.effective_edns, self.effective_edns_max_udp_payload, self.effective_edns_flags, self.effective_edns_options, self.effective_tcp)
+        clone = DNSResponse(self.message, self.msg_size, self.error, self.errno, self.history, self.response_time, self.query, self.server_cookie, self.server_cookie_status, review_history=False)
+        clone.set_effective_request_options(self.effective_flags, self.effective_edns, self.effective_edns_max_udp_payload, self.effective_edns_flags, self.effective_edns_options, self.effective_tcp, self.effective_server_cookie_status)
         clone.set_responsiveness(self.udp_attempted, self.udp_responsive, self.tcp_attempted, self.tcp_responsive, self.responsive_cause_index, self.responsive_cause_index_tcp)
         return clone
 
-    def set_effective_request_options(self, flags, edns, edns_max_udp_payload, edns_flags, edns_options, effective_tcp):
+    def set_effective_request_options(self, flags, edns, edns_max_udp_payload, edns_flags, edns_options, tcp, server_cookie_status):
         self.effective_flags = flags
         self.effective_edns = edns
         self.effective_edns_max_udp_payload = edns_max_udp_payload
         self.effective_edns_flags = edns_flags
         self.effective_edns_options = edns_options
-        self.effective_tcp = effective_tcp
+        self.effective_tcp = tcp
+        self.effective_server_cookie_status = server_cookie_status
 
     def set_responsiveness(self, udp_attempted, udp_responsive, tcp_attempted, tcp_responsive, responsive_cause_index, responsive_cause_index_tcp):
         self.udp_attempted = udp_attempted
@@ -181,7 +324,8 @@ class DNSResponse:
         edns = self.query.edns
         edns_max_udp_payload = self.query.edns_max_udp_payload
         edns_flags = self.query.edns_flags
-        edns_options = self.query.edns_options[:]
+        edns_options = copy.deepcopy(self.query.edns_options)
+        server_cookie_status = self.server_cookie_status
 
         # mark whether TCP or UDP was attempted initially
         tcp_attempted = tcp = self.query.tcp
@@ -257,10 +401,16 @@ class DNSResponse:
                 filtered_options = [x for x in edns_options if retry.action_arg == x.otype]
                 if filtered_options:
                     edns_options.remove(filtered_options[0])
+                    # If COOKIE option was removed, then reset
+                    # server_cookie_status
+                    if filtered_options[0].otype == 10:
+                        server_cookie_status = Q.DNS_COOKIE_NO_COOKIE
             elif retry.action == Q.RETRY_ACTION_CHANGE_SPORT:
                 pass
             elif retry.action == Q.RETRY_ACTION_CHANGE_EDNS_VERSION:
                 edns = retry.action_arg
+            elif retry.action == Q.RETRY_ACTION_UPDATE_DNS_COOKIE:
+                server_cookie_status = Q.DNS_COOKIE_SERVER_COOKIE_FRESH
 
             prev_index = i
 
@@ -286,7 +436,14 @@ class DNSResponse:
                     responsive_cause_index = prev_index
                     responsive_cause_index_tcp = tcp
 
-        self.set_effective_request_options(flags, edns, edns_max_udp_payload, edns_flags, edns_options, tcp)
+        # If EDNS was effectively disabled, reset EDNS options
+        if edns < 0:
+            edns_max_udp_payload = None
+            edns_flags = 0
+            edns_options = []
+            server_cookie_status = Q.DNS_COOKIE_NO_COOKIE
+
+        self.set_effective_request_options(flags, edns, edns_max_udp_payload, edns_flags, edns_options, tcp, server_cookie_status)
         self.set_responsiveness(udp_attempted, udp_responsive, tcp_attempted, tcp_responsive, responsive_cause_index, responsive_cause_index_tcp)
 
     def recursion_desired(self):
@@ -334,7 +491,7 @@ class DNSResponse:
 
         return self.message is not None and bool(self.message.flags & dns.flags.AA)
 
-    def is_referral(self, qname, rdtype, bailiwick, proper=False):
+    def is_referral(self, qname, rdtype, rdclass, bailiwick, proper=False):
         '''Return True if this response yields a referral for the queried
         name.'''
 
@@ -350,25 +507,25 @@ class DNSResponse:
             return False
         # if the name exists in the answer section with the requested rdtype or
         # CNAME, then it can't be a referral
-        if [x for x in self.message.answer if x.name == qname and x.rdtype in (rdtype, dns.rdatatype.CNAME)]:
+        if [x for x in self.message.answer if x.name == qname and x.rdtype in (rdtype, dns.rdatatype.CNAME) and x.rdclass == rdclass]:
             return False
         # if an SOA record with the given qname exists, then the server
         # is authoritative for the name, so it is a referral
         try:
-            self.message.find_rrset(self.message.authority, qname, dns.rdataclass.IN, dns.rdatatype.SOA)
+            self.message.find_rrset(self.message.authority, qname, rdclass, dns.rdatatype.SOA)
             return False
         except KeyError:
             pass
         # if proper referral is requested and qname is equal to of an NS RRset
         # in the authority, then it is a referral
         if proper:
-            if [x for x in self.message.authority if qname == x.name and x.rdtype == dns.rdatatype.NS]:
+            if [x for x in self.message.authority if qname == x.name and x.rdtype == dns.rdatatype.NS and x.rdclass == rdclass]:
                 return True
         # if proper referral is NOT requested, qname is a subdomain of
         # (including equal to) an NS RRset in the authority, and qname is not
         # equal to bailiwick, then it is a referral
         else:
-            if [x for x in self.message.authority if qname.is_subdomain(x.name) and bailiwick != x.name and x.rdtype == dns.rdatatype.NS]:
+            if [x for x in self.message.authority if qname.is_subdomain(x.name) and bailiwick != x.name and x.rdtype == dns.rdatatype.NS and x.rdclass == rdclass]:
                 return True
         return False
 
@@ -510,7 +667,7 @@ class DNSResponse:
             for o in self.effective_edns_options:
                 s = io.BytesIO()
                 o.to_wire(s)
-                d['effective_query_options']['edns_options'].append(base64.b64encode(s.getvalue()))
+                d['effective_query_options']['edns_options'].append((o.type, binascii.hexlify(s.getvalue())))
             d['effective_query_options']['tcp'] = self.effective_tcp
 
             if self.responsive_cause_index is not None:
@@ -543,7 +700,7 @@ class DNSResponse:
         return d
 
     @classmethod
-    def deserialize(cls, d, query):
+    def deserialize(cls, d, query, server_cookie, server_cookie_status):
         from . import query as Q
 
         if 'msg_size' in d:
@@ -590,7 +747,7 @@ class DNSResponse:
         history = []
         for retry in d['history']:
             history.append(Q.DNSQueryRetryAttempt.deserialize(retry))
-        return DNSResponse(message, msg_size, error, errno1, history, response_time, query)
+        return DNSResponse(message, msg_size, error, errno1, history, response_time, query, server_cookie, server_cookie_status)
 
 class DNSResponseComponent(object):
     def __init__(self):
@@ -731,6 +888,10 @@ class DNSKEYMeta(DNSResponseComponent):
         elif rdata.algorithm in (13,14):
             return len(key_str)<<3
 
+        # EDDSA keys
+        elif rdata.algorithm in (15,16):
+            return len(key_str)<<3
+
         # other keys - just guess, based on the length of the raw key material
         else:
             return len(key_str)<<3
@@ -761,7 +922,7 @@ class DNSKEYMeta(DNSResponseComponent):
         d = OrderedDict()
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -873,7 +1034,7 @@ class RRsetInfo(DNSResponseComponent):
             rdata_wire = rdata.to_digestable()
             rdata_len = len(rdata_wire)
 
-            stuff = struct.pack("!HHIH", rrset.rdtype, rrset.rdclass,
+            stuff = struct.pack(b'!HHIH', rrset.rdtype, rrset.rdclass,
                                 ttl, rdata_len)
             s += name_wire + stuff + rdata_wire
 
@@ -882,27 +1043,27 @@ class RRsetInfo(DNSResponseComponent):
     def get_rrsig_info(self, rrsig):
         return self.rrsig_info[rrsig]
 
-    def update_rrsig_info(self, server, client, response, section, is_referral):
+    def update_rrsig_info(self, server, client, response, section, rdclass, is_referral):
         try:
-            rrsig_rrset = response.message.find_rrset(section, self.rrset.name, dns.rdataclass.IN, dns.rdatatype.RRSIG, self.rrset.rdtype)
+            rrsig_rrset = response.message.find_rrset(section, self.rrset.name, rdclass, dns.rdatatype.RRSIG, self.rrset.rdtype)
             for rrsig in rrsig_rrset:
-                self.create_or_update_rrsig_info(rrsig, rrsig_rrset.ttl, server, client, response, is_referral)
+                self.create_or_update_rrsig_info(rrsig, rrsig_rrset.ttl, server, client, response, rdclass, is_referral)
         except KeyError:
             pass
 
         if self.dname_info is not None:
-            self.dname_info.update_rrsig_info(server, client, response, section, is_referral)
+            self.dname_info.update_rrsig_info(server, client, response, section, rdclass, is_referral)
 
-    def create_or_update_rrsig_info(self, rrsig, ttl, server, client, response, is_referral):
+    def create_or_update_rrsig_info(self, rrsig, ttl, server, client, response, rdclass, is_referral):
         try:
             rrsig_info = self.get_rrsig_info(rrsig)
         except KeyError:
             rrsig_info = self.rrsig_info[rrsig] = RDataMeta(self.rrset.name, ttl, dns.rdatatype.RRSIG, rrsig)
         rrsig_info.add_server_client(server, client, response)
-        self.set_wildcard_info(rrsig, server, client, response, is_referral)
+        self.set_wildcard_info(rrsig, server, client, response, rdclass, is_referral)
 
-    def create_or_update_cname_from_dname_info(self, synthesized_cname_info, server, client, response):
-        return self.insert_into_list(synthesized_cname_info, self.cname_info_from_dname, server, client, response)
+    def create_or_update_cname_from_dname_info(self, synthesized_cname_info, server, client, response, rdclass):
+        return self.insert_into_list(synthesized_cname_info, self.cname_info_from_dname, server, client, response, rdclass)
 
     def is_wildcard(self, rrsig):
         if self.rrset.name[0] == '*':
@@ -914,13 +1075,13 @@ class RRsetInfo(DNSResponseComponent):
             return dns.name.Name(('*',)+self.rrset.name.labels[-(rrsig.labels+1):])
         return self.rrset.name
 
-    def set_wildcard_info(self, rrsig, server, client, response, is_referral):
+    def set_wildcard_info(self, rrsig, server, client, response, rdclass, is_referral):
         if self.is_wildcard(rrsig):
             wildcard_name = self.reduce_wildcard(rrsig)
             if wildcard_name not in self.wildcard_info:
                 self.wildcard_info[wildcard_name] = NegativeResponseInfo(self.rrset.name, self.rrset.rdtype, self.ttl_cmp)
             self.wildcard_info[wildcard_name].add_server_client(server, client, response)
-            self.wildcard_info[wildcard_name].create_or_update_nsec_info(server, client, response, is_referral)
+            self.wildcard_info[wildcard_name].create_or_update_nsec_info(server, client, response, rdclass, is_referral)
 
     def message_for_rrsig(self, rrsig):
 
@@ -941,7 +1102,7 @@ class RRsetInfo(DNSResponseComponent):
         d = OrderedDict()
 
         if html_format:
-            formatter = lambda x: cgi.escape(x, True)
+            formatter = lambda x: escape(x, True)
         else:
             formatter = lambda x: x
 
@@ -1006,10 +1167,10 @@ class NegativeResponseInfo(DNSResponseComponent):
     def __hash__(self):
         return hash(id(self))
 
-    def create_or_update_soa_info(self, server, client, response, is_referral):
-        soa_rrsets = [x for x in response.message.authority if x.rdtype == dns.rdatatype.SOA and self.qname.is_subdomain(x.name)]
+    def create_or_update_soa_info(self, server, client, response, rdclass, is_referral):
+        soa_rrsets = [x for x in response.message.authority if x.rdtype == dns.rdatatype.SOA and x.rdclass == rdclass and self.qname.is_subdomain(x.name)]
         if not soa_rrsets:
-            soa_rrsets = [x for x in response.message.authority if x.rdtype == dns.rdatatype.SOA]
+            soa_rrsets = [x for x in response.message.authority if x.rdtype == dns.rdatatype.SOA and x.rdclass == rdclass]
         soa_rrsets.sort(reverse=True)
         try:
             soa_rrset = soa_rrsets[0]
@@ -1021,13 +1182,13 @@ class NegativeResponseInfo(DNSResponseComponent):
 
         soa_rrset_info = RRsetInfo(soa_rrset, self.ttl_cmp)
         soa_rrset_info = self.insert_into_list(soa_rrset_info, self.soa_rrset_info, server, client, response)
-        soa_rrset_info.update_rrsig_info(server, client, response, response.message.authority, is_referral)
+        soa_rrset_info.update_rrsig_info(server, client, response, response.message.authority, rdclass, is_referral)
 
         return soa_rrset_info
 
-    def create_or_update_nsec_info(self, server, client, response, is_referral):
+    def create_or_update_nsec_info(self, server, client, response, rdclass, is_referral):
         for rdtype in dns.rdatatype.NSEC, dns.rdatatype.NSEC3:
-            nsec_rrsets = [x for x in response.message.authority if x.rdtype == rdtype]
+            nsec_rrsets = [x for x in response.message.authority if x.rdtype == rdtype and x.rdclass == rdclass]
             if not nsec_rrsets:
                 continue
 
@@ -1035,7 +1196,7 @@ class NegativeResponseInfo(DNSResponseComponent):
             nsec_set_info = self.insert_into_list(nsec_set_info, self.nsec_set_info, server, client, response)
 
             for name in nsec_set_info.rrsets:
-                nsec_set_info.rrsets[name].update_rrsig_info(server, client, response, response.message.authority, is_referral)
+                nsec_set_info.rrsets[name].update_rrsig_info(server, client, response, response.message.authority, rdclass, is_referral)
 
 class NSECSet(DNSResponseComponent):
     def __init__(self, rrsets, referral, ttl_cmp):
@@ -1103,8 +1264,8 @@ class NSECSet(DNSResponseComponent):
         for name, rrset_info in self.rrsets.items():
             rrset_info.add_server_client(server, client, response)
 
-    def create_or_update_rrsig_info(self, name, rrsig, ttl, server, client, response, is_referral):
-        self.rrsets[name].create_or_update_rrsig_info(rrsig, ttl, server, client, response, is_referral)
+    def create_or_update_rrsig_info(self, name, rrsig, ttl, server, client, response, rdclass, is_referral):
+        self.rrsets[name].create_or_update_rrsig_info(rrsig, ttl, server, client, response, rdclass, is_referral)
 
     def is_valid_nsec3_name(self, nsec_name, algorithm):
         # python3/python2 dual compatibility
